"""Test Request."""

from __future__ import annotations

from copy import copy
from pathlib import Path

import pytest


async def test_base(receive, send):
    from asgi_tools import Request

    # Request is lazy
    request = Request({}, receive, send)
    assert request is not None

    scope = {
        "type": "http",
        "asgi": {"version": "3.0"},
        "http_version": "1.1",
        "method": "GET",
        "headers": [
            (b"host", b"testserver:8000"),
            (b"accept", b"*/*"),
            (b"accept-encoding", b"gzip, deflate"),
            (b"connection", b"keep-alive"),
            (b"content-type", b"application/x-www-form-urlencoded"),
            (b"user-agent", b"python-httpx/0.16.1"),
            (b"test-header", b"test-value"),
            (b"cookie", b"session=test-session"),
        ],
        "scheme": "http",
        "path": "/testurl",
        "query_string": b"a=1%202",
        "server": ("testserver", 8000),
        "client": ("127.0.0.1", 123),
        "root_path": "",
    }

    async def receive2():
        return {"body": b"name=test%20passed"}

    request = Request(scope, receive2, send)
    assert request.method == "GET"
    assert request.headers
    assert request.headers["User-Agent"] == "python-httpx/0.16.1"
    assert request.url
    assert str(request.url) == "http://testserver:8000/testurl?a=1%202"
    assert request.client == ("127.0.0.1", 123)
    assert request.cookies
    assert request.cookies["session"] == "test-session"
    assert request.http_version == "1.1"
    assert request.type == "http"
    assert request["type"] == "http"

    formdata = await request.form()
    assert formdata
    assert formdata["name"] == "test passed"

    with pytest.raises(RuntimeError):
        assert await request.body()

    r2 = copy(request)
    assert r2 is not request


async def test_media(gen_request):
    req = gen_request()
    assert req.media
    assert not req.content_type

    req = gen_request(headers={"content-type": "text/html; charset=iso-8859-1"})
    assert req.media
    assert req.media["charset"]
    assert req.media["content_type"]
    assert req.content_type == "text/html"


async def test_body(gen_request):
    req = gen_request(body=[b"any"])
    body = await req.body()
    assert body == b"any"

    body = await req.body()
    assert body == b"any"


async def test_json(gen_request):
    from asgi_tools.errors import ASGIDecodeError

    req = gen_request(body=[b"invalid"])
    with pytest.raises(ASGIDecodeError):
        await req.json()

    req = gen_request(body=[b'{"test": 42}'])
    json = await req.json()
    assert json == {"test": 42}


async def test_data(client_cls, gen_request):
    from asgi_tools import Request, ResponseMiddleware

    async def simple_app(scope, receive, send):
        request = Request(scope, receive, send)
        data = await request.data()
        return (isinstance(data, (str, bytes)) and data) or dict(data)  # type: ignore[]

    app = ResponseMiddleware(simple_app)
    client = client_cls(app)

    # Post formdata
    res = await client.post("/", data={"test": "passed"})
    assert res.status_code == 200
    assert await res.json() == {"test": "passed"}

    # Post json
    res = await client.post("/", json={"test": "passed"})
    assert res.status_code == 200
    assert await res.json() == {"test": "passed"}

    # Post other
    res = await client.post("/", data="test passed")
    assert res.status_code == 200
    assert await res.text() == "test passed"

    # Invalid data
    res = await client.post(
        "/",
        data="invalid",
        headers={"content-type": "application/json"},
    )
    assert res.status_code == 200
    assert await res.text() == "invalid"

    req = gen_request(body=[b"invalid"], headers={"content-type": "application/json"})
    assert await req.data() == b"invalid"

    from asgi_tools.errors import ASGIDecodeError

    with pytest.raises(ASGIDecodeError):
        await req.data(raise_errors=True)


async def test_multipart(client_cls):
    from asgi_tools import Request, ResponseHTML

    async def app(scope, receive, send):
        request = Request(scope, receive, send)
        data = await request.form()
        response = ResponseHTML(data["test"].read().decode().split("\n")[0])
        return await response(scope, receive, send)

    client = client_cls(app)
    with Path(__file__).open() as f:
        res = await client.post("/", data={"test": f})
    assert res.status_code == 200
    assert await res.text() == '"""Test Request."""'
    assert res.headers["content-length"] == str(len('"""Test Request."""'))


async def test_multipart_max_size(client_cls):
    from asgi_tools import Request, ResponseHTML

    async def app(scope, receive, send):
        request = Request(scope, receive, send)
        data = await request.form(max_size=100)
        if data:
            response = ResponseHTML(data["test"].read().decode().split("\n")[0])
        else:
            response = ResponseHTML("No data")
        return await response(scope, receive, send)

    client = client_cls(app)
    with Path(__file__).open() as f:
        res = await client.post("/", data={"test": f})
    assert res.status_code == 200
    body = await res.text()
    assert await res.text() == "No data"
    assert res.headers["content-length"] == str(len("No data"))


# ruff: noqa: N803



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Mapping,
    MutableMapping,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from .request import Request

TASGIMessage = Mapping[str, Any]
TASGISend = Callable[[TASGIMessage], Awaitable[None]]
TASGIReceive = Callable[[], Awaitable[TASGIMessage]]
TASGIScope = MutableMapping[str, Any]
TASGIHeaders = list[tuple[bytes, bytes]]
TASGIApp = Callable[[TASGIScope, TASGIReceive, TASGISend], Awaitable[Any]]

TJSON = Union[None, bool, int, float, str, list["TJSON"], Mapping[str, "TJSON"]]
TExceptionHandler = Callable[["Request", BaseException], Coroutine[None, None, Any]]

TV = TypeVar("TV")
TVCallable = TypeVar("TVCallable", bound=Callable)
TVAsyncCallable = TypeVar("TVAsyncCallable", bound=Callable[..., Coroutine])
TVExceptionHandler = TypeVar("TVExceptionHandler", bound=TExceptionHandler)



"""ASGI Tools Responses Tests."""

from __future__ import annotations

from http import cookies
from typing import TYPE_CHECKING, List

import pytest

if TYPE_CHECKING:
    from asgi_tools.types import TASGIMessage


async def test_response():
    from asgi_tools import Response

    response = Response("Content", content_type="text/html", cookies={"lang": "en"})
    response.cookies["session"] = "test-session"
    response.cookies["session"]["path"] = "/"
    assert response.status_code == 200
    assert response.content == b"Content"

    messages = await read_response(response)
    assert messages
    assert messages[0] == {
        "headers": [
            (b"content-type", b"text/html; charset=utf-8"),
            (b"content-length", b"7"),
            (b"set-cookie", b"lang=en"),
            (b"set-cookie", b"session=test-session; Path=/"),
        ],
        "status": 200,
        "type": "http.response.start",
    }
    assert messages[1] == {"body": b"Content", "type": "http.response.body"}

    response = Response(b"image", content_type="image/png")
    messages = await read_response(response)
    assert messages == [
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [(b"content-type", b"image/png"), (b"content-length", b"5")],
        },
        {"type": "http.response.body", "body": b"image"},
    ]


async def test_html_response():
    from asgi_tools import ResponseHTML

    response = ResponseHTML("Content")
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/html; charset=utf-8"
    assert response.content == b"Content"

    messages = await read_response(response)
    assert messages == [
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"text/html; charset=utf-8"),
                (b"content-length", b"7"),
            ],
        },
        {"type": "http.response.body", "body": b"Content"},
    ]


async def test_text_response():
    from asgi_tools import ResponseText

    response = ResponseText("Content")
    assert response.status_code == 200
    assert response.headers["content-type"] == "text/plain; charset=utf-8"
    assert response.content == b"Content"

    messages = await read_response(response)
    assert messages == [
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"text/plain; charset=utf-8"),
                (b"content-length", b"7"),
            ],
        },
        {"type": "http.response.body", "body": b"Content"},
    ]


async def test_json_response():
    from asgi_tools import ResponseJSON

    response = ResponseJSON([1, 2, 3])
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    assert response.content == b"[1,2,3]"

    messages = await read_response(response)
    assert messages == [
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"application/json"),
                (b"content-length", b"7"),
            ],
        },
        {"type": "http.response.body", "body": b"[1,2,3]"},
    ]

    response = ResponseJSON(None)
    messages = await read_response(response)
    assert messages == [
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"application/json"),
                (b"content-length", b"4"),
            ],
        },
        {"type": "http.response.body", "body": b"null"},
    ]


async def test_redirect_response():
    from asgi_tools import ResponseRedirect

    response = ResponseRedirect("/logout")
    assert response.status_code == 307
    assert response.headers["location"] == "/logout"
    assert response.content == b""
    messages = await read_response(response)
    assert messages == [
        {
            "type": "http.response.start",
            "status": 307,
            "headers": [(b"location", b"/logout"), (b"content-length", b"0")],
        },
        {"type": "http.response.body", "body": b""},
    ]


async def test_error_response():
    from asgi_tools import ResponseError

    response = ResponseError(status_code=503)
    assert response.content == b"The server cannot process the request due to a high load"

    response = ResponseError.NOT_FOUND()
    assert response.status_code == 404
    assert response.content == b"Nothing matches the given URI"

    response = ResponseError.INTERNAL_SERVER_ERROR("custom message")
    assert response.status_code == 500
    assert response.content == b"custom message"


# TODO: Exceptions
async def test_stream_response(client_cls):
    from asgi_tools import ResponseStream
    from asgi_tools._compat import aio_sleep

    async def filler(timeout=0.001):
        for idx in range(10):
            await aio_sleep(timeout)
            yield idx

    response = ResponseStream(filler())
    messages = await read_response(response)
    assert len(messages) == 12
    assert messages[-2] == {
        "body": b"9",
        "more_body": True,
        "type": "http.response.body",
    }
    assert messages[-1] == {"body": b"", "type": "http.response.body"}

    def app(scope, receive, send):
        response = ResponseStream(filler())
        return response(scope, receive, send)

    client = client_cls(app)
    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "0123456789"


async def test_file_response():
    from asgi_tools import ASGIError, ResponseFile

    response = ResponseFile(__file__)
    assert response.headers["content-length"]
    assert response.headers["content-type"] == "text/x-python"
    assert response.headers["last-modified"]
    assert response.headers["etag"]
    assert "content-disposition" not in response.headers

    response = ResponseFile(__file__, filename="tests.py")
    assert response.headers["content-disposition"] == "attachment; filename*=UTF-8''tests.py"

    messages = await read_response(response)
    assert len(messages) >= 3
    assert b"ASGI Tools Responses Tests" in messages[1]["body"]

    response = ResponseFile(__file__, headers_only=True)
    messages = await read_response(response)
    assert len(messages) == 3
    assert messages[1] == {
        "type": "http.response.body",
        "body": b"",
        "more_body": True,
    }

    with pytest.raises(ASGIError):
        response = ResponseFile("unknown")


async def test_sse_response(client_cls):
    from asgi_tools import ResponseSSE
    from asgi_tools._compat import aio_sleep

    async def filler(timeout=0.001):
        for _idx in range(2):
            await aio_sleep(timeout)
            yield "data: test"
            yield {"event": "ping"}

    response = ResponseSSE(filler())
    messages = await read_response(response)
    assert messages[1]["body"] == b"data: test\n\n"
    assert messages[2]["body"] == b"event: ping\n\n"

    def app(scope, receive, send):
        response = ResponseSSE(filler())
        return response(scope, receive, send)

    client = client_cls(app)
    res = await client.get("/")
    assert res.status_code == 200
    text = await res.text()
    assert "data: test\n\n" in text
    assert "event: ping\n\n" in text


async def test_websocket_response(client_cls):
    import json

    from asgi_tools import ASGIConnectionClosedError, ResponseWebSocket
    from asgi_tools.tests import Pipe

    pipe = Pipe()

    ws = ResponseWebSocket({}, pipe.receive_from_app, pipe.send_to_client)
    await pipe.send_to_app({"type": "websocket.connect"})
    await ws.accept()
    msg = await pipe.receive_from_client()
    assert msg == {"type": "websocket.accept"}
    await pipe.send_to_app({"type": "websocket.disconnect"})
    msg = await ws.receive()
    assert msg == {"type": "websocket.disconnect"}
    assert not ws.connected

    async def app(scope, receive, send):
        async with ResponseWebSocket(scope, receive, send) as ws:
            await ws.accept()
            msg = await ws.receive()
            assert msg == "ping"
            await ws.send("pong")
            await ws.send_json(["ping", "pong"])

    async with client_cls(app).websocket("/") as ws:
        await ws.send("ping")
        msg = await ws.receive()
        assert msg == "pong"
        msg = json.loads(await ws.receive())
        assert msg == ["ping", "pong"]
        with pytest.raises(ASGIConnectionClosedError):
            await ws.receive()


async def test_parse_response():
    from asgi_tools import parse_response

    response = parse_response({"test": "passed"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json"
    _, body = await read_response(response)
    assert body == {"body": b'{"test":"passed"}', "type": "http.response.body"}

    response = parse_response((500, "SERVER ERROR"))
    assert response.status_code == 500
    assert response.content == b"SERVER ERROR"

    response = parse_response((302, {"location": "https://google.com"}, "go away"))
    assert response.status_code == 302
    assert response.content == b"go away"
    assert response.headers["location"] == "https://google.com"

    with pytest.raises(AssertionError):
        parse_response((None, "SERVER ERROR"))


async def read_response(response) -> List[TASGIMessage]:
    from functools import partial

    from asgi_tools._compat import aio_sleep
    from asgi_tools.utils import to_awaitable

    messages = []
    await response(None, partial(aio_sleep, 10), to_awaitable(messages.append))
    return messages


# ruff: noqa: N803



"""Simple Base for ASGI Apps."""

from __future__ import annotations

from functools import partial
from inspect import isclass
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

from http_router import PrefixedRoute

from asgi_tools.router import Router

from .errors import ASGIConnectionClosedError, ASGIInvalidMethodError, ASGINotFoundError
from .logs import logger
from .middleware import LifespanMiddleware, StaticFilesMiddleware, parse_response
from .request import Request
from .response import Response, ResponseError
from .utils import iscoroutinefunction, to_awaitable

if TYPE_CHECKING:
    import logging
    from pathlib import Path

    from http_router.types import TMethods, TPath

    from .types import (
        TASGIReceive,
        TASGIScope,
        TASGISend,
        TExceptionHandler,
        TVExceptionHandler,
    )

T = TypeVar("T")
TRouteHandler = TypeVar("TRouteHandler", bound=Callable[[Request], Any])
TCallable = TypeVar("TCallable", bound=Callable[..., Any])


class App:
    """A helper class to build ASGI applications quickly and efficiently.

    Features:
        - Routing with flexible path and method matching
        - ASGI-Tools :class:`Request`, :class:`Response` integration
        - Exception management and custom error handlers
        - Static files serving
        - Lifespan events (startup/shutdown)
        - Simple middleware support

    :param debug: Enable debug mode (more logging, raise unhandled exceptions)
    :type debug: bool
    :param logger: Custom logger for the application
    :type logger: logging.Logger
    :param static_url_prefix: URL prefix for static files
    :type static_url_prefix: str
    :param static_folders: List of folders to serve static files from
    :type static_folders: Optional[list[str | Path]]
    :param trim_last_slash: Treat "/path" and "/path/" as the same route
    :type trim_last_slash: bool
    :raises ValueError: If static_url_prefix is set without static_folders
    :raises TypeError: If static_folders is not a list of strings or Paths
    :raises RuntimeError: If the app is not an ASGI application

    Example:
        >>> from asgi_tools import App
        >>> app = App()
        >>> @app.route("/")
        ... async def homepage(request):
        ...     return "Hello, World!"
    """

    exception_handlers: dict[type[BaseException], TExceptionHandler]

    def __init__(
        self,
        *,
        debug: bool = False,
        logger: logging.Logger = logger,
        static_url_prefix: str = "/static",
        static_folders: list[str | Path] | None = None,
        trim_last_slash: bool = False,
    ):
        """Initialize the ASGI application with routing, logging, static files,
           and lifespan support.

        :param debug: Enable debug mode (more logging, raise unhandled exceptions)
        :type debug: bool
        :param logger: Custom logger for the application
        :type logger: logging.Logger
        :param static_url_prefix: URL prefix for static files
        :type static_url_prefix: str
        :param static_folders: List of folders to serve static files from
        :type static_folders: Optional[list[str|Path]]
        :param trim_last_slash: Treat "/path" and "/path/" as the same route
        :type trim_last_slash: bool
        """

        # Register base exception handlers
        self.exception_handlers = {
            ASGIConnectionClosedError: to_awaitable(lambda _, __: None),
        }

        # Setup routing
        self.router = Router(
            trim_last_slash=trim_last_slash, validator=callable, converter=to_awaitable
        )

        # Setup logging
        self.logger = logger

        # Setup app
        self.__app__ = self.__match__

        # Setup lifespan
        self.lifespan = LifespanMiddleware(
            self.__process__, ignore_errors=not debug, logger=self.logger
        )

        # Enable middleware for static files
        if static_folders and static_url_prefix:
            md = StaticFilesMiddleware.setup(folders=static_folders, url_prefix=static_url_prefix)
            self.middleware(md)

        # Debug mode
        self.debug = debug

        # Handle unknown exceptions
        if not debug:

            async def handle_unknown_exception(_: Request, exc: BaseException) -> Response:
                self.logger.exception(exc)
                return ResponseError.INTERNAL_SERVER_ERROR()

            self.exception_handlers[Exception] = handle_unknown_exception

        self.internal_middlewares: list = []

    async def __call__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend) -> None:
        """ASGI entrypoint. Converts the given scope into a request and processes it."""
        await self.lifespan(scope, receive, send)

    async def __process__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Internal request processing: builds a Request, calls the app, and handles exceptions."""
        scope["app"] = self
        request = Request(scope, receive, send)
        try:
            response: Response | None = await self.__app__(request, receive, send)
            if response is not None:
                await response(scope, receive, send)

        # Handle exceptions
        except BaseException as exc:
            for etype in type(exc).mro():
                if etype in self.exception_handlers:
                    await parse_response(
                        await self.exception_handlers[etype](request, exc),
                    )(scope, receive, send)
                    break
            else:
                if isinstance(exc, Response):
                    await exc(scope, receive, send)
                else:
                    raise

    async def __match__(
        self, request: Request, _: TASGIReceive, send: TASGISend
    ) -> Response | None:
        """Finds and calls a route handler, parses the response, and handles routing exceptions."""
        scope = request.scope
        path = f"{ scope.get('root_path', '') }{ scope['path'] }"
        try:
            match = self.router(path, scope.get("method", "GET"))

        except ASGINotFoundError as exc:
            raise ResponseError.NOT_FOUND() from exc

        except ASGIInvalidMethodError as exc:
            raise ResponseError.METHOD_NOT_ALLOWED() from exc

        scope["endpoint"] = match.target
        scope["path_params"] = {} if match.params is None else match.params
        handler = cast("Callable[[Request], Any]", match.target)
        response = await handler(request)

        if scope["type"] == "http":
            return parse_response(response)

        # TODO: Do we need to close websockets automatically?
        # if scope_type == "websocket" send websocket.close

        return None

    def __route__(self, router: Router, *prefixes: str, **_) -> "App":
        """Mount this app as a nested application under given prefixes."""
        for prefix in prefixes:
            route = RouteApp(prefix, set(), target=self)
            router.dynamic.insert(0, route)
        return self

    def middleware(self, md: TCallable, *, insert_first: bool = False) -> TCallable:
        """Register a middleware for the application.

        Middleware can be a coroutine (request/response) or a regular callable (lifespan).

        :param md: The middleware function or class
        :type md: TCallable
        :param insert_first: If True, insert as the first middleware
        :type insert_first: bool
        :return: The registered middleware
        :rtype: TCallable

        Example:
            >>> @app.middleware
            ... async def log_requests(handler, request):
            ...     print(f"Request: {request.method} {request.url}")
            ...     return await handler(request)
        """
        # Register as a simple middleware
        if iscoroutinefunction(md):
            if md not in self.internal_middlewares:
                if insert_first:
                    self.internal_middlewares.insert(0, md)
                else:
                    self.internal_middlewares.append(md)

            app = self.__match__
            for md_ in reversed(self.internal_middlewares):
                app = partial(md_, app)

            self.__app__ = app

        else:
            self.lifespan.bind(md(self.lifespan.app))

        return md

    def route(
        self, *paths: TPath, methods: TMethods | None = None, **opts: Any
    ) -> Callable[[TRouteHandler], TRouteHandler]:
        """Register a route handler.

        :param paths: One or more URL paths to match
        :type paths: TPath
        :param methods: HTTP methods to match (GET, POST, etc.)
        :type methods: Optional[TMethods]
        :param opts: Additional options for the route
        :type opts: Any
        :return: Decorator function
        :rtype: Callable[[TRouteHandler], TRouteHandler]
        """
        return self.router.route(*paths, methods=methods, **opts)

    def on_startup(self, fn: Callable) -> None:
        """Register a startup event handler.

        :param fn: The function to call on startup
        :type fn: Callable
        """
        return self.lifespan.on_startup(fn)

    def on_shutdown(self, fn: Callable) -> None:
        """Register a shutdown event handler.

        :param fn: The function to call on shutdown
        :type fn: Callable
        """
        return self.lifespan.on_shutdown(fn)

    def on_error(self, etype: type[BaseException]):
        """Register a custom exception handler for a given exception type.

        :param etype: The exception type to handle
        :type etype: type[BaseException]
        :return: A decorator to register the handler
        :rtype: Callable

        Example:
            >>> @app.on_error(TimeoutError)
            ... async def timeout_handler(request, error):
            ...     return 'Timeout occurred'
        """
        assert isclass(etype), f"Invalid exception type: {etype}"
        assert issubclass(etype, BaseException), f"Invalid exception type: {etype}"

        def recorder(handler: TVExceptionHandler) -> TVExceptionHandler:
            self.exception_handlers[etype] = to_awaitable(handler)
            return handler

        return recorder


class RouteApp(PrefixedRoute):
    """Custom route to submount an application under a given path prefix."""

    def __init__(self, path: str, methods: set, target: App):
        """Create a submounted app callable for the given prefix."""
        path = path.rstrip("/")

        def app(request: Request):
            subrequest = request.__copy__(path=request.path[len(path) :])
            return target.__app__(subrequest, request.receive, request.send)

        super().__init__(path, methods, app)



"""Work with multipart."""

from __future__ import annotations

from io import BytesIO
from tempfile import SpooledTemporaryFile
from typing import TYPE_CHECKING, Callable
from urllib.parse import unquote_to_bytes

from multidict import MultiDict

from .multipart import BaseParser, MultipartParser, QueryStringParser
from .utils import parse_options_header

if TYPE_CHECKING:
    from asgi_tools.request import Request


async def read_formdata(
    request: "Request",
    max_size: int,
    upload_to: Callable | None,
    file_memory_limit: int = 1024 * 1024,
) -> MultiDict:
    """Read formdata from the given request."""
    if request.content_type == "multipart/form-data":
        reader: FormReader = MultipartReader(
            request.charset,
            upload_to,
            file_memory_limit,
        )
    else:
        reader = FormReader(request.charset)

    parser = reader.init_parser(request, max_size)
    async for chunk in request.stream():
        parser.write(chunk)

    parser.finalize()
    return reader.form


class FormReader:
    """Process querystring form data."""

    __slots__ = "charset", "curname", "curvalue", "form"

    def __init__(self, charset: str):
        self.charset = charset
        self.curname = bytearray()
        self.curvalue = bytearray()
        self.form: MultiDict = MultiDict()

    def init_parser(self, _: "Request", max_size: int) -> BaseParser:
        return QueryStringParser(
            {
                "field_name": self.on_field_name,
                "field_data": self.on_field_data,
                "field_end": self.on_field_end,
            },
            max_size=max_size,
        )

    def on_field_name(self, data: bytes, start: int, end: int):
        self.curname += data[start:end]

    def on_field_data(self, data: bytes, start: int, end: int):
        self.curvalue += data[start:end]

    def on_field_end(self, *_):
        self.form.add(
            unquote_plus(self.curname).decode(self.charset),
            unquote_plus(self.curvalue).decode(self.charset),
        )
        self.curname.clear()
        self.curvalue.clear()


class MultipartReader(FormReader):
    """Process multipart formdata."""

    __slots__ = (
        "charset",
        "curname",
        "curvalue",
        "file_memory_limit",
        "form",
        "headers",
        "name",
        "partdata",
        "upload_to",
    )

    def __init__(self, charset: str, upload_to: Callable | None, file_memory_limit: int):
        super().__init__(charset)
        self.name = ""
        self.headers: dict[bytes, bytes] = {}
        self.partdata = BytesIO()
        self.upload_to = upload_to
        self.file_memory_limit = file_memory_limit

    def init_parser(self, request: "Request", max_size: int) -> BaseParser:
        boundary = request.media.get("boundary", "")
        if not boundary:
            raise ValueError("Invalid content type boundary")

        return MultipartParser(
            boundary,
            {
                "header_end": self.on_header_end,
                "header_field": self.on_header_field,
                "headers_finished": self.on_headers_finished,
                "header_value": self.on_header_value,
                "part_data": self.on_part_data,
                "part_end": self.on_part_end,
            },
            max_size=max_size,
        )

    def on_header_field(self, data: bytes, start: int, end: int):
        self.curname += data[start:end]

    def on_header_value(self, data: bytes, start: int, end: int):
        self.curvalue += data[start:end]

    def on_header_end(self, *_):
        self.headers[bytes(self.curname.lower())] = bytes(self.curvalue)
        self.curname.clear()
        self.curvalue.clear()

    def on_headers_finished(self, *_):
        _, options = parse_options_header(
            self.headers[b"content-disposition"].decode(self.charset),
        )
        self.name = options["name"]
        if "filename" in options:
            upload_to = self.upload_to
            if upload_to is not None:
                filename = upload_to(options["filename"])
                self.partdata = f = open(filename, "wb+")  # noqa: SIM115, PTH123

            else:
                self.partdata = f = SpooledTemporaryFile(self.file_memory_limit)  # noqa: SIM115
                f._file.name = options["filename"]  # type: ignore[]

            f.content_type = self.headers[b"content-type"].decode(self.charset)

    def on_part_data(self, data: bytes, start: int, end: int):
        self.partdata.write(data[start:end])

    def on_part_end(self, *_):
        field_data = self.partdata
        if isinstance(field_data, BytesIO):
            self.form.add(self.name, field_data.getvalue().decode(self.charset))

        else:
            field_data.seek(0)
            self.form.add(self.name, field_data)

        self.partdata = BytesIO()
        self.headers = {}


def unquote_plus(value: bytearray) -> bytes:
    value = value.replace(b"+", b" ")
    return unquote_to_bytes(bytes(value))



from __future__ import annotations


def test_parse_options_header():
    from asgi_tools.utils import parse_options_header

    ct, opts = parse_options_header('text/plain; charset="utf-8"')
    assert ct == "text/plain"
    assert opts == {"charset": "utf-8"}

    ct, opts = parse_options_header(
        'form-data; name="test_client.py"; filename="test_client.py"',
    )
    assert ct == "form-data"
    assert opts == {"name": "test_client.py", "filename": "test_client.py"}


async def test_awaitable():
    from asgi_tools.utils import is_awaitable, to_awaitable

    def test1():
        return 1

    async def test2():
        return 2

    async def test3():
        yield 3

    assert not is_awaitable(test1)
    assert is_awaitable(test2)
    assert is_awaitable(test3)

    assert await to_awaitable(test1)() == 1



"""ASGI-Tools Middlewares."""

from __future__ import annotations

import abc
from contextlib import suppress
from contextvars import ContextVar
from functools import partial
from inspect import isawaitable
from pathlib import Path
from typing import TYPE_CHECKING, Awaitable, Callable, Final, Mapping

from http_router import Router

from .errors import ASGIError
from .logs import logger
from .request import Request
from .response import Response, ResponseError, ResponseFile, ResponseRedirect, parse_response
from .utils import to_awaitable

if TYPE_CHECKING:
    from .types import TASGIApp, TASGIMessage, TASGIReceive, TASGIScope, TASGISend


class BaseMiddleware(metaclass=abc.ABCMeta):
    """Base class for ASGI-Tools middlewares."""

    scopes: tuple[str, ...] = ("http", "websocket")

    def __init__(self, app: TASGIApp | None = None) -> None:
        """Save ASGI App."""
        self.bind(app)

    def __call__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend) -> Awaitable:
        """Handle ASGI call."""
        if scope["type"] in self.scopes:
            return self.__process__(scope, receive, send)

        return self.app(scope, receive, send)

    @abc.abstractmethod
    async def __process__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Do the middleware's logic."""
        raise NotImplementedError()

    @classmethod
    def setup(cls, **params) -> Callable:
        """Setup the middleware without an initialization."""
        return partial(cls, **params)  # type: ignore[abstract]

    def bind(self, app: TASGIApp | None = None):
        """Rebind the middleware to an ASGI application if it has been inited already."""
        self.app = app or ResponseError.NOT_FOUND()
        return self


# Backward compatibility
BaseMiddeware = BaseMiddleware


class ResponseMiddleware(BaseMiddleware):
    """Automatically convert ASGI_ apps results into responses :class:`~asgi_tools.Response` and
    send them to server as ASGI_ messages.

    .. code-block:: python

        from asgi_tools import ResponseMiddleware, ResponseText, ResponseRedirect

        async def app(scope, receive, send):
            # ResponseMiddleware catches ResponseError, ResponseRedirect and convert the exceptions
            # into HTTP response
            if scope['path'] == '/user':
                raise ResponseRedirect('/login')

            # Return ResponseHTML
            if scope['method'] == 'GET':
                return '<b>HTML is here</b>'

            # Return ResponseJSON
            if scope['method'] == 'POST':
                return {'json': 'here'}

            # Return any response explicitly
            if scope['method'] == 'PUT':
                return ResponseText('response is here')

            # Short form to responses: (status_code, body) or (status_code, body, headers)
            return 405, 'Unknown method'

        app = ResponseMiddleware(app)

    The conversion rules:

    * :class:`Response` objects will be directly returned from the view
    * ``dict``, ``list``, ``int``, ``bool``, ``None`` results will be converted
      into :class:`ResponseJSON`
    * ``str``, ``bytes`` results will be converted into :class:`ResponseHTML`
    * ``tuple[int, Any, dict]`` will be converted into a :class:`Response` with
      ``int`` status code, ``dict`` will be used as headers, ``Any`` will be used
      to define the response's type

    .. code-block:: python

        from asgi_tools import ResponseMiddleware

        # The result will be converted into HTML 404 response with the 'Not Found' body
        async def app(request, receive, send):
            return 404, 'Not Found'

        app = ResponseMiddleware(app)

    You are able to raise :class:`ResponseError` from yours ASGI_ apps and it
    will be catched and returned as a response

    """

    async def __process__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Parse responses from callbacks."""
        try:
            result = await self.app(scope, receive, self.send)
            response = parse_response(result)
            await response(scope, receive, send)

        except (ResponseError, ResponseRedirect) as exc:
            await exc(scope, receive, send)

    def send(self, _: TASGIMessage):
        raise RuntimeError("You can't use send() method in ResponseMiddleware")

    def bind(self, app: TASGIApp | None = None):
        """Rebind the middleware to an ASGI application if it has been inited already."""
        self.app = app or to_awaitable(lambda *_: ResponseError.NOT_FOUND())
        return self


class RequestMiddleware(BaseMiddleware):
    """Automatically create :class:`asgi_tools.Request` from the scope and pass it to ASGI_ apps.

    .. code-block:: python

        from asgi_tools import RequestMiddleware, Response

        async def app(scope, receive, send):
            content = f"{ request.method } { request.url.path }"
            response = Response(content)
            await response(scope, receive, send)

        app = RequestMiddleware(app)

    """

    async def __process__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Replace scope with request object."""
        return await self.app(Request(scope, receive, send), receive, send)


class LifespanMiddleware(BaseMiddleware):
    """Manage ASGI_ Lifespan events.

    :param ignore_errors: Ignore errors from startup/shutdown handlers
    :param on_startup: the list of callables to run when the app is starting
    :param on_shutdown: the list of callables to run when the app is finishing

    .. code-block:: python

        from asgi_tools import LifespanMiddleware, Response

        async def app(scope, receive, send):
            response = Response('OK')
            await response(scope, receive, send)

        app = lifespan = LifespanMiddleware(app)

        @lifespan.on_startup
        async def start():
            print('The app is starting')

        @lifespan.on_shutdown
        async def start():
            print('The app is finishing')

    Lifespan middleware may be used as an async context manager for testing purposes

    .. code-block: python

        async def test_my_app():

            # ...

            # Registered startup/shutdown handlers will be called
            async with lifespan:
                # ... do something

    """

    scopes = ("lifespan",)

    def __init__(
        self,
        app: TASGIApp | None = None,
        *,
        logger=logger,
        ignore_errors: bool = False,
        on_startup: Callable | list[Callable] | None = None,
        on_shutdown: Callable | list[Callable] | None = None,
    ) -> None:
        """Prepare the middleware."""
        super(LifespanMiddleware, self).__init__(app)
        self.ignore_errors = ignore_errors
        self.logger = logger
        self.__startup__: list[Callable] = []
        self.__shutdown__: list[Callable] = []
        self.__register__(on_startup, self.__startup__)
        self.__register__(on_shutdown, self.__shutdown__)

    async def __process__(self, _: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Manage lifespan cycle."""
        while True:
            message = await receive()
            if message["type"] == "lifespan.startup":
                msg = await self.run("startup", send)
                await send(msg)

            elif message["type"] == "lifespan.shutdown":
                msg = await self.run("shutdown", send)
                await send(msg)
                break

    def __register__(
        self, handlers: Callable | list[Callable] | None, container: list[Callable]
    ) -> None:
        """Register lifespan handlers."""
        if not handlers:
            return

        if not isinstance(handlers, list):
            handlers = [handlers]

        container += handlers

    async def __aenter__(self):
        """Use the lifespan middleware as a context manager."""
        await self.run("startup")
        return self

    async def __aexit__(self, *_):
        """Use the lifespan middleware as a context manager."""
        await self.run("shutdown")

    async def run(self, event: str, _: TASGISend | None = None):
        """Run startup/shutdown handlers."""
        assert event in {"startup", "shutdown"}
        handlers = getattr(self, f"__{event}__")

        for handler in handlers:
            try:
                res = handler()
                if isawaitable(res):
                    await res

            except Exception as exc:  # noqa: PERF203
                self.logger.exception("%s method '%s' raises an exception.", event.title(), handler)
                if self.ignore_errors:
                    continue

                self.logger.exception("Lifespans process failed")
                return {"type": f"lifespan.{event}.failed", "message": str(exc)}

        return {"type": f"lifespan.{event}.complete"}

    def on_startup(self, fn: Callable) -> None:
        """Add a function to startup."""
        self.__register__(fn, self.__startup__)

    def on_shutdown(self, fn: Callable) -> None:
        """Add a function to shutdown."""
        self.__register__(fn, self.__shutdown__)


class RouterMiddleware(BaseMiddleware):
    r"""Manage routing.

    .. code-block:: python

        from asgi_tools import RouterMiddleware, ResponseHTML, ResponseError

        async def default_app(scope, receive, send):
            response = ResponseError.NOT_FOUND()
            await response(scope, receive, send)

        app = router = RouterMiddleware(default_app)

        @router.route('/status', '/stat')
        async def status(scope, receive, send):
            response = ResponseHTML('STATUS OK')
            await response(scope, receive, send)

        # Bind methods
        # ------------
        @router.route('/only-post', methods=['POST'])
        async def only_post(scope, receive, send):
            response = ResponseHTML('POST OK')
            await response(scope, receive, send)

        # Regexp paths
        # ------------
        import re

        @router.route(re.compile(r'/\d+/?'))
        async def num(scope, receive, send):
            num = int(scope['path'].strip('/'))
            response = ResponseHTML(f'Number { num }')
            await response(scope, receive, send)

        # Dynamic paths
        # -------------

        @router.route('/hello/{name}')
        async def hello(scope, receive, send):
            name = scope['path_params']['name']
            response = ResponseHTML(f'Hello { name.title() }')
            await response(scope, receive, send)

        # Set regexp for params
        @router.route(r'/multiply/{first:\d+}/{second:\d+}')
        async def multiply(scope, receive, send):
            first, second = map(int, scope['path_params'].values())
            response = ResponseHTML(str(first * second))
            await response(scope, receive, send)

    Path parameters are made available in the request/scope, as the ``path_params`` dictionary.

    """

    def __init__(self, app: TASGIApp | None = None, router: Router | None = None) -> None:
        """Initialize HTTP router."""
        super().__init__(app)
        self.router = router or Router(validator=callable)

    async def __process__(self, scope: TASGIScope, *args):
        """Get an app and process."""
        app, scope["path_params"] = self.__dispatch__(scope)
        return await app(scope, *args)

    def __dispatch__(self, scope: TASGIScope) -> tuple[Callable, Mapping | None]:
        """Lookup for a callback."""
        path = f"{scope.get('root_path', '')}{scope['path']}"
        try:
            match = self.router(path, scope["method"])

        except self.router.RouterError:
            return self.app, {}

        else:
            return match.target, match.params  # type: ignore[]

    def route(self, *args, **kwargs):
        """Register a route."""
        return self.router.route(*args, **kwargs)


class StaticFilesMiddleware(BaseMiddleware):
    """Serve static files.

    :param url_prefix:  an URL prefix for static files
    :type url_prefix: str, "/static"
    :param folders: Paths to folders with static files
    :type folders: list[str]

    .. code-block:: python

        from asgi_tools import StaticFilesMiddleware, ResponseHTML

        async def app(scope, receive, send):
            response = ResponseHTML('OK')
            await response(scope, receive, send)

        # Files from static folder will be served from /static
        app = StaticFilesMiddleware(app, folders=['static'])

    """

    scopes = ("http",)

    def __init__(
        self,
        app: TASGIApp | None = None,
        url_prefix: str = "/static",
        folders: list[str | Path] | None = None,
    ) -> None:
        """Initialize the middleware."""
        super().__init__(app)
        self.url_prefix = url_prefix
        folders = folders or []
        self.folders: list[Path] = [Path(folder) for folder in folders]

    async def __process__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend) -> None:
        """Serve static files for self url prefix."""
        path = scope["path"]
        url_prefix = self.url_prefix
        if path.startswith(url_prefix):
            response: Response | None = None
            filename = path[len(url_prefix) :].strip("/")
            for folder in self.folders:
                filepath = folder.joinpath(filename).resolve()
                if folder != filepath.parent:
                    continue
                with suppress(ASGIError):
                    response = ResponseFile(filepath, headers_only=scope["method"] == "HEAD")
                    break

            response = response or ResponseError(status_code=404)
            await response(scope, receive, send)

        else:
            await self.app(scope, receive, send)


BACKGROUND_TASK: Final = ContextVar[list[Awaitable] | None]("background_task", default=None)


class BackgroundMiddleware(BaseMiddleware):
    """Run background tasks.


    .. code-block:: python

        from asgi_tools import BackgroundMiddleware, ResponseText

        async def app(scope, receive, send):
            response = ResponseText('OK)

            # Schedule any awaitable for later execution
            BackgroundMiddleware.set_task(asyncio.sleep(1))

            # Return response immediately
            await response(scope, receive, send)

            # The task will be executed after the response is sent

        app = BackgroundMiddleware(app)

    """

    async def __process__(self, scope: TASGIScope, receive: TASGIReceive, send: TASGISend):
        """Run background tasks."""
        await self.app(scope, receive, send)
        for task in BACKGROUND_TASK.get() or []:
            await task

        BACKGROUND_TASK.set(None)  # Clear the context variable

    @staticmethod
    def set_task(task: Awaitable):
        """Set a task for background execution."""
        tasks = BACKGROUND_TASK.get() or []
        tasks.append(task)
        BACKGROUND_TASK.set(tasks)



"""Testing tools."""

from __future__ import annotations

import asyncio
import binascii
import io
import mimetypes
import os
import random
from collections import deque
from contextlib import asynccontextmanager, suppress
from functools import partial
from http.cookies import SimpleCookie
from json import loads
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    Deque,
    cast,
)
from urllib.parse import urlencode

from yarl import URL

from ._compat import aio_cancel, aio_sleep, aio_spawn, aio_timeout, aio_wait
from .constants import BASE_ENCODING, DEFAULT_CHARSET
from .errors import ASGIConnectionClosedError, ASGIInvalidMessageError
from .response import Response, ResponseJSON, ResponseWebSocket, parse_websocket_msg
from .utils import CIMultiDict, parse_headers

if TYPE_CHECKING:
    from multidict import MultiDict

    from .types import TJSON, TASGIApp, TASGIMessage, TASGIReceive, TASGIScope, TASGISend


class TestResponse(Response):
    """Response for test client."""

    def __init__(self):
        super().__init__(b"")
        self.content = None

    async def __call__(self, _: TASGIScope, receive: TASGIReceive, send: TASGISend):  # noqa: ARG002
        self._receive = receive
        msg = await self._receive()
        assert msg.get("type") == "http.response.start", "Invalid Response"
        self.status_code = int(msg.get("status", 502))
        self.headers = cast("MultiDict", parse_headers(msg.get("headers", [])))
        self.content_type = self.headers.get("content-type")
        for cookie in self.headers.getall("set-cookie", []):
            self.cookies.load(cookie)

    async def stream(self) -> AsyncGenerator[bytes, None]:
        """Stream the response."""
        more_body = True
        while more_body:
            msg = await self._receive()
            if msg.get("type") == "http.response.body":
                chunk = msg.get("body")
                if chunk:
                    yield chunk
                more_body = msg.get("more_body", False)

    async def body(self) -> bytes:
        """Load response body."""
        if self.content is None:
            body = b""
            async for chunk in self.stream():
                body += chunk
            self.content = body

        return self.content

    async def text(self) -> str:
        body = await self.body()
        return body.decode(DEFAULT_CHARSET)

    async def json(self) -> TJSON:
        text = await self.text()
        return loads(text)


class TestWebSocketResponse(ResponseWebSocket):
    """Support websockets in tests."""

    def connect(self) -> Coroutine[TASGIMessage, Any, Any]:
        return self.send({"type": "websocket.connect"})

    async def disconnect(self):
        await self.send({"type": "websocket.disconnect", "code": 1005})
        self.state = self.STATES.DISCONNECTED

    def send(self, msg, msg_type="websocket.receive"):
        """Send a message to a client."""
        return super().send(msg, msg_type=msg_type)

    async def receive(self, *, raw=False):
        """Receive messages from a client."""
        if self.partner_state == self.STATES.DISCONNECTED:
            raise ASGIConnectionClosedError

        msg = await self._receive()
        if not msg["type"].startswith("websocket."):
            raise ASGIInvalidMessageError(msg)

        if msg["type"] == "websocket.accept":
            self.partner_state = self.STATES.CONNECTED
            return await self.receive(raw=raw)

        if msg["type"] == "websocket.close":
            self.partner_state = self.STATES.DISCONNECTED
            raise ASGIConnectionClosedError

        return msg if raw else parse_websocket_msg(msg, charset=DEFAULT_CHARSET)


class ASGITestClient:
    """Built-in test client for ASGI applications.

    Features:

    * cookies
    * multipart/form-data
    * follow redirects
    * request streams
    * response streams
    * websocket support
    * lifespan management

    """

    def __init__(self, app: TASGIApp, base_url: str = "http://localhost"):
        self.app = app
        self.base_url = URL(base_url)
        self.cookies: SimpleCookie = SimpleCookie()
        self.headers: dict[str, str] = {}

    def __getattr__(self, name: str) -> Callable[..., Awaitable]:
        return partial(self.request, method=name.upper())

    async def request(
        self,
        path: str,
        method: str = "GET",
        *,
        query: str | dict = "",
        headers: dict[str, str] | None = None,
        cookies: dict[str, str] | None = None,
        data: bytes | str | dict | AsyncGenerator[Any, bytes] = b"",
        json: TJSON = None,
        follow_redirect: bool = True,
        timeout: float = 10.0,
    ) -> TestResponse:
        """Make a HTTP requests."""

        headers = headers or dict(self.headers)

        if isinstance(data, str):
            data = Response.process_content(data)

        elif isinstance(data, dict):
            is_multipart = any(isinstance(value, io.IOBase) for value in data.values())
            if is_multipart:
                data, headers["Content-Type"] = encode_multipart(data)

            else:
                headers["Content-Type"] = "application/x-www-form-urlencoded"
                data = urlencode(data).encode(DEFAULT_CHARSET)

        elif json is not None:
            headers["Content-Type"] = "application/json"
            data = ResponseJSON.process_content(json)

        pipe = Pipe()

        if isinstance(data, bytes):
            headers.setdefault("Content-Length", str(len(data)))

        scope = self.build_scope(
            path,
            type="http",
            query=query,
            method=method,
            headers=headers,
            cookies=cookies,
        )

        async with aio_timeout(timeout):
            await aio_wait(
                pipe.stream(data),
                self.app(scope, pipe.receive_from_app, pipe.send_to_client),
            )

        res = TestResponse()
        await res(scope, pipe.receive_from_client, pipe.send_to_app)
        for n, v in res.cookies.items():
            self.cookies[n] = v

        if follow_redirect and res.status_code in {301, 302, 303, 307, 308}:
            return await self.get(res.headers["location"])

        return res

    # TODO: Timeouts for websockets
    @asynccontextmanager
    async def websocket(
        self,
        path: str,
        query: str | dict | None = None,
        headers: dict | None = None,
        cookies: dict | None = None,
    ):
        """Connect to a websocket."""
        pipe = Pipe()

        ci_headers = CIMultiDict(headers or {})

        scope = self.build_scope(
            path,
            headers=ci_headers,
            query=query,
            cookies=cookies,
            type="websocket",
            subprotocols=str(ci_headers.get("Sec-WebSocket-Protocol", "")).split(","),
        )
        ws = TestWebSocketResponse(scope, pipe.receive_from_client, pipe.send_to_app)
        async with aio_spawn(
            self.app,
            scope,
            pipe.receive_from_app,
            pipe.send_to_client,
        ):
            await ws.connect()
            yield ws
            await ws.disconnect()

    def lifespan(self, timeout: float = 3e-2):
        """Manage `Lifespan <https://asgi.readthedocs.io/en/latest/specs/lifespan.html>`_
        protocol."""
        return manage_lifespan(self.app, timeout=timeout)

    def build_scope(
        self,
        path: str,
        headers: dict | CIMultiDict | None = None,
        query: str | dict | None = None,
        cookies: dict | None = None,
        **scope,
    ) -> TASGIScope:
        """Prepare a request scope."""
        headers = headers or {}
        headers.setdefault("User-Agent", "ASGI-Tools-Test-Client")
        headers.setdefault("Host", self.base_url.host)

        if cookies:
            for c, v in cookies.items():
                self.cookies[c] = v

        if len(self.cookies):
            headers.setdefault("Cookie", self.cookies.output(header="", sep=";"))

        url = URL(path)
        if query:
            url = url.with_query(query)

        # Setup client
        scope.setdefault("client", ("127.0.0.1", random.randint(1024, 65535)))  # noqa: S311

        return dict(
            {
                "asgi": {"version": "3.0"},
                "http_version": "1.1",
                "path": url.path,
                "query_string": url.raw_query_string.encode(),
                "raw_path": url.raw_path.encode(),
                "root_path": "",
                "scheme": (scope.get("type") == "http" and self.base_url.scheme) or "ws",
                "headers": [
                    (key.lower().encode(BASE_ENCODING), str(val).encode(BASE_ENCODING))
                    for key, val in (headers or {}).items()
                ],
                "server": ("127.0.0.1", self.base_url.port),
            },
            **scope,
        )


def encode_multipart(data: dict) -> tuple[bytes, str]:
    body = io.BytesIO()
    boundary = binascii.hexlify(os.urandom(16))
    for name, data_value in data.items():
        value = data_value
        headers = f'Content-Disposition: form-data; name="{ name }"'
        if hasattr(value, "read"):
            filename = getattr(value, "name", None)
            if filename:
                headers = f'{ headers }; filename="{ Path(filename).name }"'
                content_type = mimetypes.guess_type(filename)[0] or "application/octet-stream"
                headers = f"{ headers }\r\nContent-Type: { content_type }"
            value = value.read()

        body.write(b"--%b\r\n" % boundary)
        body.write(headers.encode("utf-8"))
        body.write(b"\r\n\r\n")
        if isinstance(value, str):
            value = value.encode("utf-8")
        body.write(value)
        body.write(b"\r\n")

    body.write(b"--%b--\r\n" % boundary)
    return body.getvalue(), (b"multipart/form-data; boundary=%s" % boundary).decode()


class Pipe:
    __slots__ = (
        "app_is_closed",
        "app_queue",
        "client_is_closed",
        "client_queue",
        "delay",
    )

    def __init__(self, delay: float = 1e-3):
        self.delay = delay
        self.app_is_closed = False
        self.client_is_closed = False
        self.app_queue: Deque[TASGIMessage] = deque()
        self.client_queue: Deque[TASGIMessage] = deque()

    async def send_to_client(self, msg: TASGIMessage):
        if self.client_is_closed:
            raise ASGIInvalidMessageError(msg.get("type"))

        if msg.get("type") == "websocket.close":
            self.client_is_closed = True

        elif msg.get("type") == "http.response.body":
            self.client_is_closed = not msg.get("more_body", False)

        self.client_queue.append(msg)

    async def send_to_app(self, msg: TASGIMessage):
        if self.app_is_closed:
            raise ASGIInvalidMessageError(msg.get("type"))

        if msg.get("type") == "http.disconnect":
            self.app_is_closed = True

        self.app_queue.append(msg)

    async def receive_from_client(self):
        while not self.client_queue:
            await aio_sleep(self.delay)
        return self.client_queue.popleft()

    async def receive_from_app(self):
        while not self.app_queue:
            await aio_sleep(self.delay)
        return self.app_queue.popleft()

    async def stream(self, data: bytes | AsyncGenerator[Any, bytes]):
        if isinstance(data, bytes):
            return await self.send_to_app(
                {"type": "http.request", "body": data, "more_body": False},
            )

        async for chunk in data:
            await self.send_to_app({"type": "http.request", "body": chunk, "more_body": True})
        await self.send_to_app({"type": "http.request", "body": b"", "more_body": False})
        return None


@asynccontextmanager
async def manage_lifespan(app, timeout: float = 3e-2):
    """Manage `Lifespan <https://asgi.readthedocs.io/en/latest/specs/lifespan.html>`_ protocol."""
    pipe = Pipe()

    scope = {"type": "lifespan"}

    async def safe_spawn():
        with suppress(BaseException):
            await app(scope, pipe.receive_from_app, pipe.send_to_client)

    async with aio_spawn(safe_spawn) as task:
        await pipe.send_to_app({"type": "lifespan.startup"})

        with suppress(TimeoutError, asyncio.TimeoutError):  # python 310
            async with aio_timeout(timeout):
                msg = await pipe.receive_from_client()
                if msg["type"] == "lifespan.startup.failed":
                    await aio_cancel(task)

        yield

        await pipe.send_to_app({"type": "lifespan.shutdown"})
        with suppress(TimeoutError, asyncio.TimeoutError):  # python 310
            async with aio_timeout(timeout):
                await pipe.receive_from_client()



"""ASGI-Tools -- Tools to make ASGI Applications"""

from __future__ import annotations

from http_router import InvalidMethodError, NotFoundError

from .app import App
from .errors import ASGIConnectionClosedError, ASGIError, ASGIInvalidMethodError, ASGINotFoundError
from .middleware import (
    BackgroundMiddleware,
    LifespanMiddleware,
    RequestMiddleware,
    ResponseMiddleware,
    RouterMiddleware,
    StaticFilesMiddleware,
)
from .request import Request
from .response import (
    Response,
    ResponseError,
    ResponseFile,
    ResponseHTML,
    ResponseJSON,
    ResponseRedirect,
    ResponseSSE,
    ResponseStream,
    ResponseText,
    ResponseWebSocket,
    parse_response,
)
from .view import HTTPView

__all__ = (  # noqa: RUF022
    # Errors
    "ASGIConnectionClosedError",
    "ASGIError",
    "ASGIInvalidMethodError",
    "ASGINotFoundError",
    "InvalidMethodError",
    "NotFoundError",
    # App/handlers
    "App",
    "HTTPView",
    # Request/Response
    "Request",
    "Response",
    "ResponseError",
    "ResponseFile",
    "ResponseHTML",
    "ResponseJSON",
    "ResponseRedirect",
    "ResponseSSE",
    "ResponseStream",
    "ResponseText",
    "ResponseWebSocket",
    # Middleware
    "BackgroundMiddleware",
    "LifespanMiddleware",
    "RequestMiddleware",
    "ResponseMiddleware",
    "RouterMiddleware",
    "StaticFilesMiddleware",
    # Utils
    "parse_response",
)



from __future__ import annotations


def test_types_available():
    from asgi_tools import types

    assert types.TASGIMessage
    assert types.TASGIScope
    assert types.TASGISend
    assert types.TASGIReceive
    assert types.TASGIApp
    assert types.TJSON
    assert types.TV
    assert types.TVCallable
    assert types.TVAsyncCallable
    assert types.TExceptionHandler
    assert types.TVExceptionHandler



from __future__ import annotations

import re


async def test_readme_request_response(client_cls):
    # Example
    import json

    from asgi_tools import Request, Response

    async def app(scope, receive, send):
        if scope["type"] != "http":
            return None

        # Parse scope
        request = Request(scope, receive, send)
        request_data = {
            # Get full URL
            "url": str(request.url),
            "charset": request.charset,
            # Get headers
            "headers": {**request.headers},
            # Get query params
            "query": dict(request.query),
            # Get cookies
            "cookies": dict(request.cookies),
        }

        # Create a response (HTMLResponse, PlainTextResponse, JSONResponse, StreamResponse,
        # RedirectResponse also available)
        response = Response(json.dumps(request_data), content_type="application/json")

        # Send ASGI messages
        return await response(scope, receive, send)

    # Test
    client = client_cls(app)

    res = await client.get("/test?var=42")
    assert res.status_code == 200
    data = await res.json()
    assert data["url"] == "http://localhost/test?var=42"
    assert data["query"] == {"var": "42"}


async def test_readme_request_response_middleware(client_cls):
    # Example
    from asgi_tools import RequestMiddleware, ResponseHTML

    async def simple_app(request, receive, send):
        # We will get a parsed request here
        data = await request.json()
        response = ResponseHTML(data["name"])
        return await response(request, receive, send)

    app = RequestMiddleware(simple_app)

    # Test
    client = client_cls(app)
    res = await client.post("/test", json={"name": "passed"})
    assert res.status_code == 200
    assert await res.text() == "passed"

    # Example
    from asgi_tools import ResponseMiddleware

    async def simple_app2(request, receive, send):
        return "Hello World!"

    app = ResponseMiddleware(simple_app2)

    # Test
    client = client_cls(app)
    res = await client.post("/test", json={"name": "passed"})
    assert res.status_code == 200
    assert res.headers["content-type"] == "text/html; charset=utf-8"
    assert await res.text() == "Hello World!"


async def test_readme_router_middleware():
    from http_router import Router

    router = Router()

    @router.route("/page1")
    async def page1(request, receive, send):
        return "page1"

    @router.route("/page2")
    async def page2(request, receive, send):
        return "page2"


async def test_docs(client_cls):
    from asgi_tools import App

    app = App()

    @app.route("/")
    async def hello_world(request):
        return "<p>Hello, World!</p>"

    client = client_cls(app)

    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "<p>Hello, World!</p>"


async def test_docs_response_redirect(client_cls):
    from asgi_tools import ResponseRedirect

    async def simple_app(scope, receive, send):
        response = ResponseRedirect("/login")
        await response(scope, receive, send)

    client = client_cls(simple_app)

    res = await client.get("/", follow_redirect=False)
    assert res.status_code == 307
    assert res.headers["location"] == "/login"

    from asgi_tools import Request, ResponseMiddleware

    async def simple_app2(scope, receive, send):
        request = Request(scope, receive, send)
        if not request.headers.get("authorization"):
            raise ResponseRedirect("/login")

        return "OK"

    app = ResponseMiddleware(simple_app2)
    client = client_cls(app)

    res = await client.get("/", follow_redirect=False)
    assert res.status_code == 307
    assert res.headers["location"] == "/login"

    res = await client.get(
        "/",
        follow_redirect=False,
        headers={"authorization": "user"},
    )
    assert res.status_code == 200
    assert await res.text() == "OK"


async def test_docs_response_error(client_cls):
    from asgi_tools import ResponseError

    async def simple_app(scope, receive, send):
        response = ResponseError("Timeout", 502)
        await response(scope, receive, send)

    client = client_cls(simple_app)

    res = await client.get("/")
    assert res.status_code == 502
    assert await res.text() == "Timeout"

    from asgi_tools import Request, ResponseError, ResponseMiddleware

    async def simple_app2(scope, receive, send):
        request = Request(scope, receive, send)
        if not request.method == "POST":
            raise ResponseError("Invalid request data", 400)

        return "OK"

    app = ResponseMiddleware(simple_app2)
    client = client_cls(app)

    res = await client.get("/")
    assert res.status_code == 400
    assert await res.text() == "Invalid request data"

    res = await client.post("/")
    assert res.status_code == 200


async def test_docs_response_stream(client_cls):
    from asgi_tools import ResponseStream
    from asgi_tools._compat import aio_sleep  # for compatibility with different async libs

    async def stream_response():
        for number in range(10):
            await aio_sleep(1e-2)
            yield str(number)

    async def app(scope, receive, send):
        response = ResponseStream(stream_response(), content_type="plain/text")

        await response(scope, receive, send)

    client = client_cls(app)

    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "0123456789"


async def test_docs_response_middleware(client_cls):
    from asgi_tools import ResponseMiddleware, ResponseRedirect, ResponseText

    async def simple_app(scope, receive, send):
        # ResponseMiddleware catches ResponseError, ResponseRedirect and convert the exceptions
        # into HTTP response
        if scope["path"] == "/user":
            raise ResponseRedirect("/login")

        # Return ResponseHTML
        if scope["method"] == "GET":
            return "<b>HTML is here</b>"

        # Return ResponseJSON
        if scope["method"] == "POST":
            return {"json": "here"}

        # Return any response explicitly
        if scope["method"] == "PUT":
            return ResponseText("response is here")

        # Short form to responses: (status_code, body) or (status_code, body, headers)
        return 405, "Unknown method"

    app = ResponseMiddleware(simple_app)
    client = client_cls(app)

    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "<b>HTML is here</b>"

    res = await client.post("/")
    assert res.status_code == 200
    assert await res.json() == {"json": "here"}

    res = await client.put("/")
    assert res.status_code == 200
    assert res.headers["content-type"] == "text/plain; charset=utf-8"
    assert await res.text() == "response is here"

    res = await client.get("/user", follow_redirect=False)
    assert res.status_code == 307

    res = await client.delete("/")
    assert res.status_code == 405


async def test_docs_router_middleware(client_cls):
    from asgi_tools import ResponseError, ResponseHTML, RouterMiddleware

    async def default_app(scope, receive, send):
        response = ResponseError.NOT_FOUND()
        await response(scope, receive, send)

    app = router = RouterMiddleware(default_app)

    @router.route("/status")
    async def status(scope, receive, send):
        response = ResponseHTML("STATUS OK")
        await response(scope, receive, send)

    # Bind methods
    # ------------
    @router.route("/only-post", methods=["POST"])
    async def only_post(scope, receive, send):
        response = ResponseHTML("POST OK")
        await response(scope, receive, send)

    # Regexp paths
    # ------------

    @router.route(re.compile(r"/\d+/?"))
    async def num(scope, receive, send):
        num = int(scope["path"].strip("/"))
        response = ResponseHTML(f"Number { num }")
        await response(scope, receive, send)

    # Dynamic paths
    # -------------

    @router.route("/hello/{name}")
    async def hello(scope, receive, send):
        name = scope["path_params"]["name"]
        response = ResponseHTML(f"Hello { name.title() }")
        await response(scope, receive, send)

    # Set regexp for params
    @router.route(r"/multiply/{first:int}/{second:int}")
    async def multiply(scope, receive, send):
        first, second = map(int, scope["path_params"].values())
        response = ResponseHTML(str(first * second))
        await response(scope, receive, send)

    client = client_cls(app)

    res = await client.get("/unknown")
    assert res.status_code == 404

    res = await client.post("/status")
    assert res.status_code == 200
    assert await res.text() == "STATUS OK"

    res = await client.get("/only-post")
    assert res.status_code == 404

    res = await client.post("/only-post")
    assert res.status_code == 200
    assert await res.text() == "POST OK"

    res = await client.post("/42")
    assert res.status_code == 200
    assert await res.text() == "Number 42"

    res = await client.post("/hello/john")
    assert res.status_code == 200
    assert await res.text() == "Hello John"

    res = await client.post("/multiply/32/56")
    assert res.status_code == 200
    assert await res.text() == "1792"



"""test middlewares"""

from __future__ import annotations

import pytest


async def test_response_middleware(client_cls):
    from asgi_tools import ResponseError, ResponseMiddleware

    # Test default response
    md = ResponseMiddleware()
    client = client_cls(md)
    res = await client.get("/")
    assert res.status_code == 404
    assert await res.text() == "Nothing matches the given URI"

    async def simple_app(*args):
        return False

    app = ResponseMiddleware(simple_app)
    client = client_cls(app)
    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "false"

    async def simple_app2(*args):
        raise ResponseError.BAD_GATEWAY()

    app2 = ResponseMiddleware(simple_app2)
    client = client_cls(app2)
    res = await client.get("/")
    assert res.status_code == 502
    assert await res.text() == "Invalid responses from another server/proxy"

    async def simple_app3(*args):
        return

    app3 = ResponseMiddleware(simple_app3)
    client = client_cls(app3)
    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "null"


async def test_request_response_middlewares(client_cls):
    from asgi_tools import RequestMiddleware, ResponseMiddleware

    async def simple_app(request, receive, send):
        data = await request.json()
        first_name = data.get("first_name", "Anonymous")
        last_name = request.query.get("last_name", "Test")
        return f"Hello {first_name} {last_name} from '{ request.url.path }'"

    app = RequestMiddleware(ResponseMiddleware(simple_app))

    client = client_cls(app)
    res = await client.post(
        "/testurl?last_name=Daniels",
        json={"first_name": "Jack"},
        headers={"test-header": "test-value"},
        cookies={"session": "test-session"},
    )
    assert res.status_code == 200
    assert await res.text() == "Hello Jack Daniels from '/testurl'"
    assert res.headers["content-length"] == str(
        len("Hello Jack Daniels from '/testurl'"),
    )


async def test_lifespan_middleware(client_cls):
    from asgi_tools import LifespanMiddleware

    side_effects = []

    async def simple_app(scope, receive, send):
        return None

    app = LifespanMiddleware(
        simple_app,
        on_startup=lambda: side_effects.append("started"),
        on_shutdown=lambda: side_effects.append("finished"),
    )
    client = client_cls(app)

    async with client.lifespan():
        assert ["started"] == side_effects

    assert ["started", "finished"] == side_effects


async def test_lifespan_middleware_errors(client_cls):
    from asgi_tools import LifespanMiddleware

    side_effects = {}

    async def fail():
        raise Exception

    async def start():
        side_effects["started"] = True

    async def simple_app(scope, receive, send):
        return None

    app = LifespanMiddleware(
        simple_app,
        on_startup=[fail, start],
        on_shutdown=lambda: side_effects.setdefault("finished", True),
    )
    client = client_cls(app)

    async with client.lifespan():
        assert "started" not in side_effects

    assert "finished" not in side_effects

    app = LifespanMiddleware.setup(ignore_errors=True)(
        lambda scope, receive, send: None,
        on_startup=[fail, start],
        on_shutdown=lambda: side_effects.setdefault("finished", True),
    )
    client = client_cls(app)

    async with client.lifespan():
        assert side_effects["started"]

    assert side_effects["finished"]


async def test_router_middleware(client_cls):
    from asgi_tools import Response, RouterMiddleware

    app = RouterMiddleware()

    with pytest.raises(app.router.RouterError):
        app.route("/test")("something")

    @app.route("/page1")
    async def page1(scope, receive, send):
        res = Response("page1")
        return await res(scope, receive, send)

    @app.route("/page2/{mode}", methods=["POST"])
    async def page2(scope, receive, send):
        mode = scope["path_params"]["mode"]
        res = Response(f"page2: {mode}")
        return await res(scope, receive, send)

    client = client_cls(app)
    res = await client.get("/")
    assert res.status_code == 404
    assert await res.text() == "Nothing matches the given URI"

    res = await client.get("/page1")
    assert res.status_code == 200
    assert await res.text() == "page1"

    res = await client.get("/page2/42")
    assert res.status_code == 404  # Returns default application

    res = await client.post("/page2/42")
    assert res.status_code == 200
    assert await res.text() == "page2: 42"


async def test_router_middleware2(client_cls):
    from asgi_tools import ResponseError, ResponseMiddleware, RouterMiddleware

    async def page404(scope, receive, send):
        return ResponseError.NOT_FOUND()

    router = RouterMiddleware(page404)
    app = ResponseMiddleware(router)

    @router.route("/page1")
    async def page1(scope, receive, send):
        return "page1"

    client = client_cls(app)
    res = await client.get("/")
    assert res.status_code == 404
    assert await res.text() == "Nothing matches the given URI"

    res = await client.get("/page1")
    assert res.status_code == 200
    assert await res.text() == "page1"


async def test_staticfiles_middleware(client_cls, app):
    from pathlib import Path

    from asgi_tools import StaticFilesMiddleware

    app = StaticFilesMiddleware(app, folders=["/", Path(__file__).parent])

    client = client_cls(app)
    res = await client.get("/")
    assert res.status_code == 200
    body = await res.body()
    assert body == b"OK"

    res = await client.head("/static/test_middlewares.py")
    assert res.status_code == 200
    assert res.headers["content-type"] == "text/x-python"
    assert not await res.text()

    res = await client.get("/static/test_middlewares.py")
    assert res.status_code == 200
    text = await res.text()
    assert text.startswith('"""test middlewares"""')

    res = await client.get("/static/unknown")
    assert res.status_code == 404

    res = await client.get("/static")
    assert res.status_code == 404


async def test_background_middleware(client_cls, app):
    import time

    from asgi_tools import BackgroundMiddleware, ResponseText, RouterMiddleware
    from asgi_tools._compat import aio_sleep

    router = RouterMiddleware()
    app = BackgroundMiddleware(router)
    results = []

    async def background_task(name):
        await aio_sleep(1e-1)
        results.append(name)

    @router.route("/test")
    async def test(scope, receive, send):
        BackgroundMiddleware.set_task(background_task("test1"))

        response = ResponseText("test")
        await response(scope, receive, send)

    client = client_cls(app)
    start = time.monotonic()
    res = await client.get("/test")
    end = time.monotonic()
    assert res.status_code == 200
    assert results == ["test1"]
    assert end - start >= 0.1  # ensure background_task executed before response end
    results.clear()

    @router.route("/multi")
    async def multi(scope, receive, send):
        BackgroundMiddleware.set_task(background_task("task1"))
        BackgroundMiddleware.set_task(background_task("task2"))
        response = ResponseText("multi")
        await response(scope, receive, send)

    res = await client.get("/multi")
    assert res.status_code == 200
    assert "task1" in results
    assert "task2" in results



"""Simple Test Client."""

from __future__ import annotations

import asyncio
import io
from pathlib import Path
from sys import version_info

import pytest

if version_info < (3, 11):
    from exceptiongroup import ExceptionGroup


def test_build_scope(client):
    scope = client.build_scope("/test", query={"value": 42, "value2": "❤️"})
    assert scope == {
        "asgi": {"version": "3.0"},
        "headers": [
            (b"user-agent", b"ASGI-Tools-Test-Client"),
            (b"host", b"localhost"),
        ],
        "http_version": "1.1",
        "path": "/test",
        "query_string": b"value=42&value2=%E2%9D%A4%EF%B8%8F",
        "raw_path": b"/test",
        "root_path": "",
        "scheme": "ws",
        "server": ("127.0.0.1", 80),
        "client": ("127.0.0.1", scope["client"][1]),
    }


async def test_client(app, client):
    res = await client.get("/")
    assert res
    assert res.content_type == "text/html; charset=utf-8"
    assert res.status_code == 200
    assert res.headers
    assert res.headers["content-type"]
    assert res.headers["content-length"] == "2"
    text = await res.text()
    assert text == "OK"

    @app.route("/test")
    async def test(request):
        data = await request.data()
        if isinstance(data, bytes):
            data = data.decode("utf-8")

        return {
            "query": dict(request.query),
            "headers": {**request.headers},
            "cookies": dict(request.cookies),
            "data": data,
        }

    res = await client.patch("/test?value=👋")
    assert res.status_code == 200
    json = await res.json()
    assert json == {
        "query": {"value": "👋"},
        "cookies": {},
        "data": "",
        "headers": {
            "host": "localhost",
            "content-length": "0",
            "user-agent": "ASGI-Tools-Test-Client",
        },
    }

    res = await client.patch("/test", data="test")
    json = await res.json()
    assert json["data"] == "test"

    res = await client.patch("/test", json={"var": 42})
    json = await res.json()
    assert json["data"] == {"var": 42}

    res = await client.patch("/test?var1=42&var2=34")
    json = await res.json()
    assert json["query"] == {"var1": "42", "var2": "34"}

    res = await client.patch("/test?var1=42&var2=34", query="var=42")
    json = await res.json()
    assert json["query"] == {"var": "42"}

    res = await client.patch("/test?var1=42&var2=34", query={"var": 42})
    json = await res.json()
    assert json["query"] == {"var": "42"}

    res = await client.patch("/test", cookies={"var": "42"})
    assert res.status_code == 200
    json = await res.json()
    assert json["cookies"] == {"var": "42"}

    # Custom methods
    # --------------
    @app.route("/caldav", methods="PROPFIND")
    async def propfind(_):
        return "PROPFIND"

    res = await client.propfind("/caldav")
    assert res.status_code == 200
    assert await res.text() == "PROPFIND"


async def test_client_response(client, app):
    @app.route("/")
    async def index(request):
        return "OK"

    res = await client.get("/")
    body = await res.body()
    assert body == b"OK"

    body = await res.body()
    assert body == b"OK"


async def test_formdata(app, client):
    @app.route("/formdata")
    async def formdata(request):
        formdata = await request.form()
        return dict(formdata)

    res = await client.post("/formdata", data={"field": "value", "field2": "value2"})
    assert res.status_code == 200
    assert await res.json() == {"field": "value", "field2": "value2"}


async def test_files(app, client):
    @app.route("/files")
    async def files(request):
        formdata = await request.form()
        return formdata["test_client.py"].read()

    res = await client.post(
        "/files",
        data={"field": "value", "test_client.py": Path(__file__).open()},
    )
    assert res.status_code == 200
    assert "test_files" in await res.text()

    fakefile = io.BytesIO(b"file content")
    fakefile.name = "test_client.py"
    res = await client.post(
        "/files",
        data={"field": "value", "test_client.py": fakefile},
    )
    assert res.status_code == 200
    assert "file content" in await res.text()


async def test_cookies(app, client):
    from asgi_tools import ResponseRedirect

    @app.route("/set-cookie")
    async def set_cookie(_):
        res = ResponseRedirect("/")
        res.cookies["c1"] = "c1"
        res.cookies["c2"] = "c2"
        return res

    res = await client.get("/set-cookie", cookies={"var": "42"})
    assert res.status_code == 200
    assert {n: v.value for n, v in client.cookies.items()} == {
        "var": "42",
        "c1": "c1",
        "c2": "c2",
    }

    @app.route("/get-cookie")
    async def get_cookie(request):
        return dict(request.cookies)

    res = await client.get("/get-cookie")
    assert await res.json() == {"var": "42", "c1": "c1", "c2": "c2"}


async def test_redirects(app, client):
    from asgi_tools import ResponseRedirect

    # Follow Redirect
    # ---------------
    @app.route("/redirect")
    async def redirect(_):
        raise ResponseRedirect("/")

    res = await client.get("/redirect")
    assert res.status_code == 200
    assert await res.text() == "OK"

    res = await client.get("/redirect", follow_redirect=False)
    assert res.status_code == 307
    assert res.headers["location"] == "/"


async def test_stream_response(app, client):
    from asgi_tools import ResponseStream
    from asgi_tools._compat import aio_sleep

    async def source(timeout=0.001):
        for idx in range(10):
            await aio_sleep(timeout)
            yield idx

    @app.route("/stream")
    async def stream(_):
        return ResponseStream(source(), content_type="plain/text")

    res = await client.get("/stream")
    assert res.status_code == 200

    expected = [str(n).encode() for n in range(10)]
    async for chunk in res.stream():
        assert chunk == expected.pop(0)


async def test_stream_request(app, client):
    from asgi_tools._compat import aio_sleep

    async def source(timeout=0.001):
        for idx in range(10):
            yield bytes(idx)
            await aio_sleep(timeout)

    @app.route("/stream")
    async def stream(request):
        data = await request.data()
        assert data

    res = await client.post("/stream", data=source())
    assert res.status_code == 200


async def test_websocket(app, client_cls):
    from asgi_tools import ASGIConnectionClosedError, ResponseWebSocket

    @app.route("/websocket")
    async def websocket(request):
        assert request.subprotocols == ["ship", "done"]
        async with ResponseWebSocket(request) as ws:
            msg = await ws.receive()
            assert msg == "ping"
            await ws.send("pong")

    async with client_cls(app).websocket(
        "/websocket",
        headers={"sec-websocket-protocol": "ship,done"},
    ) as ws:
        await ws.send("ping")
        msg = await ws.receive()
        assert msg == "pong"

        with pytest.raises(ASGIConnectionClosedError):
            await ws.receive()


async def test_websocket_disconnect(app, client_cls):
    from asgi_tools import ResponseWebSocket

    @app.route("/websocket")
    async def websocket(request):
        async with ResponseWebSocket(request) as ws:
            msg = await ws.receive()
            assert msg == {"type": "websocket.disconnect", "code": 1005}

    async with client_cls(app).websocket("/websocket"):
        pass


async def test_timeouts(app, client):
    from asgi_tools._compat import aio_sleep

    @app.route("/sleep/{time}")
    async def sleep(request):
        time = float(request.path_params["time"])
        await aio_sleep(time)
        return "OK"

    res = await client.get("/sleep/0.01")
    assert res.status_code == 200
    assert await res.text() == "OK"

    with pytest.raises((TimeoutError, asyncio.TimeoutError)):  # python 39, 310
        await client.get("/sleep/10", timeout=0.1)


async def test_lifespan_unsupported(client_cls):
    from asgi_tools import Response

    async def app(scope, receive, send):
        assert scope["type"] == "http"
        await Response("OK")(scope, receive, send)

    client = client_cls(app)

    async with client.lifespan():
        res = await client.get("/")
        assert res.status_code == 200


async def test_lifespan(client_cls):
    from asgi_tools import Response

    side_effects = {"started": False, "finished": False}

    async def app(scope, receive, send):
        if scope["type"] == "lifespan":
            while True:
                msg = await receive()
                if msg["type"] == "lifespan.startup":
                    side_effects["started"] = True
                    await send({"type": "lifespan.startup.complete"})

                elif msg["type"] == "lifespan.shutdown":
                    side_effects["finished"] = True
                    await send({"type": "lifespan.shutdown.complete"})
                    return

        await Response("OK")(scope, receive, send)

    client = client_cls(app)

    with pytest.raises((AssertionError, ExceptionGroup)):
        async with client.lifespan():
            raise AssertionError("test")

    async with client.lifespan():
        assert side_effects["started"]
        assert not side_effects["finished"]
        res = await client.get("/")
        assert res.status_code == 200

    assert side_effects["started"]
    assert side_effects["finished"]


async def test_invalid_app(client_cls):
    from asgi_tools import Response
    from asgi_tools.errors import ASGIInvalidMessageError

    async def invalid(scope, receive, send):
        await Response("test")(scope, receive, send)
        await Response("test")(scope, receive, send)

    client = client_cls(invalid)
    with pytest.raises((ASGIInvalidMessageError, ExceptionGroup)):
        await client.get("/")


# ruff: noqa: N803



from __future__ import annotations


class ASGIError(Exception):
    """Base class for ASGI-Tools Errors."""


class ASGIConnectionClosedError(ASGIError):
    """ASGI-Tools connection closed error."""


class ASGIDecodeError(ASGIError, ValueError):
    """ASGI-Tools decoding error."""


class ASGINotFoundError(ASGIError):
    """Raise when http handler not found."""


class ASGIInvalidMethodError(ASGIError):
    """Raise when http method not found."""


class ASGIInvalidMessageError(ASGIError):
    """Raise when unexpected message received."""



# cython: language_level=3


cdef class BaseParser:

    cdef dict callbacks

    cdef void callback(self, str name, bytes data, int start, int end) except *

    cpdef void write(self, bytes data) except *

    cpdef void finalize(self)


cdef class QueryStringParser(BaseParser):

    cdef unsigned int cursize
    cdef unsigned int max_size
    cdef unsigned char state


cdef class MultipartParser(BaseParser):

    cdef unsigned int cursize
    cdef unsigned int max_size
    cdef unsigned char state
    cdef unsigned int index
    cdef short flags
    cdef int header_field_pos
    cdef int header_value_pos
    cdef int part_data_pos
    cdef bytes boundary
    cdef frozenset boundary_chars
    cdef list lookbehind



from __future__ import annotations

import tempfile
from pathlib import Path

import pytest
import yaml


async def test_formdata(gen_request, tmp_path):
    from asgi_tools.forms import read_formdata
    from asgi_tools.tests import encode_multipart

    request = gen_request(body=[b"answer=42&na", b"mes=bob&test", b"&names=alice"])
    formdata = await read_formdata(request, 0, None, 0)
    assert formdata["answer"] == "42"
    assert formdata["test"] == ""
    assert formdata.getall("names") == ["bob", "alice"]

    data, content_type = encode_multipart(
        {
            "file1": open(__file__),
            "file2": open(__file__),
        },
    )
    body = [chunk + b"\n" for chunk in data.split(b"\n")]
    request = gen_request(body=body, headers={"content-type": content_type})
    formdata = await read_formdata(request, 0, None, 0)
    assert formdata["file1"]
    assert formdata["file1"].name == Path(__file__).name
    assert formdata["file1"].content_type == "text/x-python"
    assert formdata["file2"]
    assert b"test_multipart_parser" in formdata["file2"].read()

    request = gen_request(body=body, headers={"content-type": content_type})
    upload_to = lambda f: f"{tmp_path}/{f}"  # noqa
    formdata = await read_formdata(request, 0, upload_to, 0)
    assert formdata["file1"]
    assert formdata["file1"].name == str(tmp_path / Path(__file__).name)
    assert formdata["file1"].content_type == "text/x-python"
    assert formdata["file2"]
    assert formdata["file2"].name.startswith(str(tmp_path))
    assert b"test_multipart_parser" in formdata["file2"].read()


@pytest.mark.parametrize(
    "sample",
    [
        (b"foo=bar", {"foo": "bar"}),
        (b"&foo=bar", {"foo": "bar"}),
        (b"foo=bar&", {"foo": "bar"}),
        (b"foo=bar&asdf=baz", {"foo": "bar", "asdf": "baz"}),
        (b"foo=bar;asdf=baz", {"foo": "bar", "asdf": "baz"}),
        (b"foo=bar&&another=asdf", {"foo": "bar", "another": "asdf"}),
        (b"foo=bar&blank&another=asdf", {"another": "asdf", "blank": "", "foo": "bar"}),
        (b"value=test%20passed", {"value": "test passed"}),
    ],
)
def test_query(sample):
    from asgi_tools.forms import FormReader

    data, expected = sample

    reader = FormReader("utf-8")
    parser = reader.init_parser(None, 0)
    feed(parser, data)
    assert dict(reader.form) == expected


DATA = Path(__file__).parent / "fixtures/multipart"


@pytest.mark.parametrize(
    "basename",
    [
        "single_field",
        "single_field_longer",
        "single_field_single_file",
        "single_field_with_leading_newlines",
        "single_file",
        "utf8_filename",
        "multiple_fields",
        "multiple_files",
        "almost_match_boundary",
        "almost_match_boundary_without_CR",
        "almost_match_boundary_without_LF",
        "almost_match_boundary_without_final_hyphen",
        "bad_initial_boundary",
        "bad_end_of_headers",
        "CR_in_header",
        "CR_in_header_value",
        "empty_header",
    ],
)
def test_multipart(basename, gen_request):
    from asgi_tools.forms import MultipartReader

    data, meta = loader(basename)

    # Initialize parsers
    reader = MultipartReader("utf-8", None, 0)
    request = gen_request(
        headers={"content-type": f"multipart/form-data; boundary={meta['boundary']}"},
    )
    parser = reader.init_parser(request, 0)
    try:
        feed(parser, data)
    except ValueError:
        assert meta["expected"]["error"]

    parser.finalize()

    for (field, data), expected in zip(reader.form.items(), meta["expected"]):
        assert field == expected["name"]
        if expected["type"] == "field":
            assert isinstance(data, str)
            assert data == expected["data"].decode("utf-8")

        else:
            assert isinstance(data, tempfile.SpooledTemporaryFile)
            assert data.name == expected["file_name"]
            data = data.read()
            assert data == expected["data"]


# Utils
# -----


def feed(parser, data):
    for n in range(0, len(data), 5):
        parser.write(data[n : n + 5])
    parser.finalize()


def loader(basename):
    data = (DATA / basename).with_suffix(".http").read_bytes()
    meta = (DATA / basename).with_suffix(".yaml")
    with meta.open("rb") as f:
        meta = yaml.load(f, Loader=yaml.SafeLoader)
    return data, meta



"""Application Tests."""
from __future__ import annotations

from pathlib import Path
from unittest import mock


async def test_app(client_cls):
    from asgi_tools.app import App

    app = App(static_folders=[Path(__file__).parent])

    @app.route("/test/{param}", methods="get")
    async def test_request(request):
        return "Done %s" % request.path_params["param"]

    client = client_cls(app)

    res = await client.get("/404")
    assert res.status_code == 404
    assert await res.text() == "Nothing matches the given URI"

    res = await client.get("/static/test_app.py")
    assert res.status_code == 200
    text = await res.text()
    assert text.startswith('"""Application Tests."""')

    res = await client.post("/test/42")
    assert res.status_code == 405
    assert await res.text() == "Specified method is invalid for this resource"

    @app.route("/data", methods="post")
    async def test_data(request):
        data = await request.data()
        return dict(data)

    res = await client.post("/data", json={"test": "passed"})
    assert res.status_code == 200
    assert await res.json() == {"test": "passed"}

    @app.route("/none")
    async def test_none(request):
        return

    res = await client.get("/none")
    assert res.status_code == 200

    @app.route("/path_params")
    async def path_params(request):
        return request["path_params"].get("unknown", 42)

    res = await client.get("/path_params")
    assert res.status_code == 200
    assert await res.text() == "42"

    @app.route("/sync")
    def sync_fn(request):
        return "Sync is ok"

    res = await client.get("/sync")
    assert res.status_code == 200
    assert await res.text() == "Sync is ok"


async def test_scope_endpoint(client_cls):
    from asgi_tools.app import App, ResponseError

    app = App()
    client = client_cls(app)

    @app.route("/")
    async def endpoint(request):
        return request.scope["endpoint"].__qualname__

    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "test_scope_endpoint.<locals>.endpoint"


async def test_errors(client_cls):
    from asgi_tools.app import App, ResponseError

    app = App()
    client = client_cls(app)

    @app.route("/502")
    async def test_response_error(request):
        raise ResponseError.BAD_GATEWAY()

    res = await client.get("/502")
    assert res.status_code == 502
    assert await res.text() == "Invalid responses from another server/proxy"

    @app.route("/error")
    async def test_unhandled_exception(request):
        raise RuntimeError("An exception")

    res = await client.get("/error")
    assert res.status_code == 500
    assert await res.text() == "Server got itself in trouble"


async def test_trim_last_slach(client_cls):
    from asgi_tools.app import App

    app = App()
    client = client_cls(app)

    @app.route("/route1")
    async def route1(request):
        return "route1"

    @app.route("/route2/")
    async def route2(request):
        return "route2"

    res = await client.get("/route1")
    assert res.status_code == 200

    res = await client.get("/route2/")
    assert res.status_code == 200

    res = await client.get("/route1/")
    assert res.status_code == 404

    res = await client.get("/route2")
    assert res.status_code == 404

    app = App(trim_last_slash=True)
    client = client_cls(app)

    @app.route("/route1")
    async def route11(request):
        return "route1"

    @app.route("/route2/")
    async def route21(request):
        return "route2"

    res = await client.get("/route1")
    assert res.status_code == 200

    res = await client.get("/route2/")
    assert res.status_code == 200

    res = await client.get("/route1/")
    assert res.status_code == 200

    res = await client.get("/route2")
    assert res.status_code == 200


async def test_app_static(client_cls):
    from asgi_tools.app import App

    app = App(static_folders=[Path(__file__).parent])
    client = client_cls(app)

    async with client.lifespan():
        res = await client.get("/static/test_app.py")
        assert res.status_code == 200
        text = await res.text()
        assert text.startswith('"""Application Tests."""')


async def test_app_handle_exception(client_cls):
    from asgi_tools.app import App, ResponseError

    app = App()

    @app.route("/500")
    async def raise_unknown(request):
        raise Exception("Unknown Exception")

    @app.route("/501")
    async def raise_response_error(request):
        res = ResponseError(status_code=501)
        raise res

    # By default we handle all exceptions as INTERNAL SERVER ERROR 500 Response
    client = client_cls(app)
    res = await client.get("/500")
    assert res.status_code == 500
    assert await res.text() == "Server got itself in trouble"

    @app.on_error(Exception)
    async def handle_unknown(request, exc):
        return "UNKNOWN: %s" % exc

    @app.on_error(ResponseError)
    async def handler(request, response):
        if response.status_code == 404:
            return "Response 404"
        return "Custom Server Error"

    async with client.lifespan():
        res = await client.get("/500")
        assert res.status_code == 200
        assert await res.text() == "UNKNOWN: Unknown Exception"

        res = await client.get("/404")
        assert res.status_code == 200
        assert await res.text() == "Response 404"

        res = await client.get("/501")
        assert res.status_code == 200
        assert await res.text() == "Custom Server Error"


async def test_app_middleware_simple(client, app):
    from asgi_tools import ResponseHTML

    md_mock = mock.MagicMock()

    @app.route("/err")
    async def err(request):
        raise RuntimeError("Handle me")

    @app.on_error(Exception)
    async def custom_exc(request, exc):
        return ResponseHTML("App Exception")

    res = await client.get("/err")
    assert res.status_code == 200
    assert await res.text() == "App Exception"

    @app.middleware
    async def first_md(app, request, receive, send):
        md_mock("first start")
        try:
            response = await app(request, receive, send)
            if "x-second-md" in response.headers:
                response.headers["x-first-md"] = response.headers["x-second-md"]
            return response
        except RuntimeError:
            return ResponseHTML("Middleware Exception")
        finally:
            md_mock("first exit")

    @app.middleware
    async def second_md(app, request, receive, send):
        md_mock("second start")
        try:
            response = await app(request, receive, send)
            response.headers["x-second-md"] = "passed"
            return response
        finally:
            md_mock("second exit")

    res = await client.get("/")
    assert res.status_code == 200
    assert res.headers["x-first-md"] == "passed"
    assert res.headers["x-second-md"] == "passed"
    assert [args[0][0] for args in md_mock.call_args_list] == [
        "first start",
        "second start",
        "second exit",
        "first exit",
    ]

    res = await client.get("/404")
    assert res.status_code == 404
    assert "x-first-md" not in res.headers
    assert "x-second-md" not in res.headers

    res = await client.get("/err")
    assert res.status_code == 200
    assert await res.text() == "Middleware Exception"


#  @pytest.mark.skip
async def test_app_middleware_classic(client, app):
    from asgi_tools import ResponseError, ResponseHTML

    @app.route("/err")
    async def err(request):
        raise RuntimeError("Handle me")

    @app.on_error(Exception)
    async def custom_exc(request, exc):
        return ResponseHTML("App Exception")

    @app.middleware
    def classic_md(app):
        async def middleware(scope, receive, send):
            headers = scope["headers"]
            auth = [v for k, v in headers if k == b"authorization"]
            if not auth:
                response = ResponseError.UNAUTHORIZED()
                await response(scope, receive, send)
                return

            await app(scope, receive, send)

        return middleware

    res = await client.get("/")
    assert res.status_code == 401

    res = await client.get("/", headers={"authorization": "any"})
    assert res.status_code == 200
    assert await res.text() == "OK"

    res = await client.get("/err", headers={"authorization": "any"})
    assert res.status_code == 200
    assert await res.text() == "App Exception"


async def test_cbv(app, client):
    from asgi_tools import HTTPView

    @app.route("/cbv")
    class Custom(HTTPView):
        async def get(self, request):
            return "CBV: get"

        async def post(self, request):
            return "CBV: post"

    res = await client.get("/cbv")
    assert res.status_code == 200
    assert await res.text() == "CBV: get"

    res = await client.post("/cbv")
    assert res.status_code == 200
    assert await res.text() == "CBV: post"

    res = await client.put("/cbv")
    assert res.status_code == 405


async def test_websockets(app, client):
    from asgi_tools import ResponseWebSocket

    @app.route("/websocket")
    async def websocket(request):
        ws = ResponseWebSocket(request)
        await ws.accept()
        msg = await ws.receive()
        assert msg == "ping"
        await ws.send("pong")
        await ws.close()

    async with client.websocket("/websocket") as ws:
        await ws.send("ping")
        msg = await ws.receive()
        assert msg == "pong"

    res = await client.get("/")
    assert res.status_code == 200


async def test_app_lifespan(app, client):
    side_effects = {}

    @app.on_startup
    def start():
        side_effects["started"] = True

    @app.on_shutdown
    def finish():
        side_effects["finished"] = True

    async with client.lifespan():
        assert side_effects["started"]
        res = await client.get("/")
        assert res.status_code == 200

    assert side_effects["finished"]


async def test_nested(app, client):
    from asgi_tools.app import App

    @app.middleware
    async def mid(app, request, receive, send):
        response = await app(request, receive, send)
        response.headers["x-app"] = "OK"
        return response

    subapp = App()

    @subapp.middleware
    async def submid(app, request, receive, send):
        response = await app(request, receive, send)
        response.headers["x-subapp"] = "OK"
        return response

    @subapp.route("/route")
    async def route(request):
        return "OK from subapp"

    app.route("/sub")(subapp)

    res = await client.get("/")
    assert res.status_code == 200
    assert await res.text() == "OK"
    assert res.headers["x-app"] == "OK"

    res = await client.get("/sub/route")
    assert res.status_code == 200
    assert await res.text() == "OK from subapp"
    assert res.headers["x-subapp"] == "OK"
    assert res.headers["x-app"] == "OK"



from __future__ import annotations

from contextlib import suppress

import pytest


@pytest.mark.benchmark(group="req-res", disable_gc=True)
def test_benchmark_req_res(benchmark, gen_request, receive):
    from asgi_tools import Request, parse_response

    scope = gen_request(
        "/test",
        query={"param": "value"},
        headers={"header": "value"},
        cookies={"cookie": "value"},
    ).scope

    async def send(msg):
        pass

    def run_benchmark():
        request = Request(scope, receive, send)
        assert request.method
        assert request.query["param"]
        assert request.headers["header"]
        assert request.cookies["cookie"]

        response = parse_response("body")
        coro = response(None, None, send)
        with suppress(StopIteration):
            coro.send(None)

    benchmark(run_benchmark)


@pytest.mark.benchmark(group="formdata", disable_gc=True)
def test_benchmark_formdata(benchmark):
    from asgi_tools.forms import FormReader

    def run_benchmark():
        reader = FormReader("utf-8")
        parser = reader.init_parser(None, 0)
        parser.write(b"&value=test%20passed")
        parser.finalize()
        return reader.form

    form = benchmark(run_benchmark)
    assert dict(form) == {"value": "test passed"}


@pytest.mark.benchmark(group="app", disable_gc=True)
def test_benchmark_app(benchmark, app, client):
    @app.route("/path")
    async def page(request):
        return "OK"

    @app.middleware
    async def md(app, request, receive, send):
        response = await app(request, receive, send)
        response.headers["x-app"] = "OK"
        return response

    scope = client.build_scope("/path", type="http")

    messages = []

    async def send(msg):
        messages.append(msg)

    def run_benchmark():
        coro = app(scope, None, send)
        with suppress(StopIteration):
            coro.send(None)

        nonlocal messages
        res = list(messages)
        messages = []
        return res

    res = benchmark(run_benchmark)
    assert res == [
        {
            "status": 200,
            "type": "http.response.start",
            "headers": [
                (b"content-type", b"text/html; charset=utf-8"),
                (b"x-app", b"OK"),
                (b"content-length", b"2"),
            ],
        },
        {
            "type": "http.response.body",
            "body": b"OK",
        },
    ]



"""Compatibility layer."""

from __future__ import annotations

import asyncio
import inspect
from asyncio import create_task, gather, sleep
from concurrent.futures import ALL_COMPLETED, FIRST_COMPLETED
from contextlib import asynccontextmanager, suppress
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    Union,
    cast,
)

if TYPE_CHECKING:
    from pathlib import Path

    from .types import TJSON


from json import dumps, loads

from sniffio import current_async_library

__all__ = (
    "aio_cancel",
    "aio_sleep",
    "aio_spawn",
    "aio_stream_file",
    "aio_wait",
    "aiofile_installed",
    "create_task",
    "curio_installed",
    "json_dumps",
    "json_loads",
    "trio_installed",
)

try:
    from asyncio import timeout as asyncio_timeout  # type: ignore[attr-defined]
except ImportError:  # python 310
    from async_timeout import timeout as asyncio_timeout  # type: ignore[no-redef]


aiofile_installed = False
with suppress(ImportError):
    import aiofile

    aiofile_installed = True


trio_installed = False
with suppress(ImportError):
    from trio import TooSlowError, open_memory_channel, open_nursery
    from trio import fail_after as trio_fail_after
    from trio import open_file as trio_open_file
    from trio import sleep as trio_sleep

    trio_installed = True


curio_installed = False
with suppress(ImportError):
    from curio import TaskGroup as CurioTaskGroup
    from curio import TaskTimeout
    from curio import aopen as curio_open
    from curio import sleep as curio_sleep
    from curio import spawn as curio_spawn
    from curio import timeout_after as curio_fail_after

    curio_installed = True


def aio_sleep(seconds: float = 0) -> Awaitable:
    """Return sleep coroutine."""

    if trio_installed and current_async_library() == "trio":
        return trio_sleep(seconds)  # noqa: ASYNC105

    if curio_installed and current_async_library() == "curio":
        return curio_sleep(seconds)

    return sleep(seconds)


@asynccontextmanager
async def aio_spawn(fn: Callable[..., Awaitable], *args, **kwargs):
    """Spawn a given coroutine."""
    if trio_installed and current_async_library() == "trio":
        async with open_nursery() as tasks:
            tasks.start_soon(fn, *args, **kwargs)
            yield tasks

    elif curio_installed and current_async_library() == "curio":
        task = await curio_spawn(fn, *args, **kwargs)
        yield task
        await task.join()  # type: ignore [union-attr]

    else:
        coro = cast("Coroutine", fn(*args, **kwargs))
        task = create_task(coro)
        yield task
        await task


@asynccontextmanager
async def aio_timeout(timeout: float):  # noqa: ASYNC109
    """Fail after the given timeout."""
    if not timeout:
        yield
        return

    if trio_installed and current_async_library() == "trio":
        try:
            with trio_fail_after(timeout):
                yield

        except TooSlowError:
            raise TimeoutError(f"{timeout}s.") from None

    elif curio_installed and current_async_library() == "curio":
        try:
            async with curio_fail_after(timeout):
                yield

        except TaskTimeout:
            raise TimeoutError(f"{timeout}s.") from None

    else:
        async with asyncio_timeout(timeout):
            yield


async def aio_wait(*aws: Awaitable, strategy: str = ALL_COMPLETED) -> Any:
    """Run the coros concurrently, wait for all completed or cancel others.

    Only ALL_COMPLETED, FIRST_COMPLETED are supported.
    """
    if not aws:
        return None

    if trio_installed and current_async_library() == "trio":
        send_channel, receive_channel = open_memory_channel(0)  # type: ignore[var-annotated]

        async with open_nursery() as n:
            for aw in aws:
                n.start_soon(trio_jockey, aw, send_channel)

            results = []
            for _ in aws:
                results.append(await receive_channel.receive())
                if strategy == FIRST_COMPLETED:
                    n.cancel_scope.cancel()
                    return results[0]

            return results

    if curio_installed and current_async_library() == "curio":
        wait = all if strategy == ALL_COMPLETED else any
        async with CurioTaskGroup(wait=wait) as g:
            [await g.spawn(aw) for aw in aws]

        return g.results if strategy == ALL_COMPLETED else g.result

    aws = tuple(create_task(aw) if inspect.iscoroutine(aw) else aw for aw in aws)
    done, pending = await asyncio.wait(aws, return_when=strategy)  # type: ignore[type-var]
    if strategy != ALL_COMPLETED:
        for task in pending:
            task.cancel()  # type: ignore[attr-defined]
        await gather(*pending, return_exceptions=True)
        return next(iter(done)).result()  # type: ignore[attr-defined]

    return [t.result() for t in done]  # type: ignore[attr-defined]


async def aio_cancel(task: Union[asyncio.Task, Any]):
    """Cancel asyncio task / trio nursery."""
    if isinstance(task, asyncio.Task):
        return task.cancel()

    if trio_installed and current_async_library() == "trio":
        return task.cancel_scope.cancel()

    if curio_installed and current_async_library() == "curio":
        return await task.cancel()
    return None


async def aio_stream_file(
    filepath: Union[str, Path], chunk_size: int = 32 * 1024
) -> AsyncGenerator[bytes, None]:
    if trio_installed and current_async_library() == "trio":
        async with await trio_open_file(filepath, "rb") as fp:
            while True:
                chunk = cast("bytes", await fp.read(chunk_size))
                if not chunk:
                    break
                yield chunk

    elif curio_installed and current_async_library() == "curio":
        async with curio_open(filepath, "rb") as fp:
            while True:
                chunk = cast("bytes", await fp.read(chunk_size))
                if not chunk:
                    break
                yield chunk

    else:
        if not aiofile_installed:
            raise RuntimeError(
                "`aiofile` is required to return files with asyncio",
            )

        async with aiofile.AIOFile(filepath, mode="rb") as fp:
            async for chunk in aiofile.Reader(  # type: ignore [assignment]
                fp, chunk_size=chunk_size
            ):
                yield cast("bytes", chunk)


async def trio_jockey(coro: Awaitable, channel):
    """Wait for the given coroutine and send result back to the given channel."""
    await channel.send(await coro)


def json_dumps(content) -> bytes:
    """Emulate orjson."""
    return dumps(  # type: ignore [call-arg]
        content,
        ensure_ascii=False,
        separators=(",", ":"),
    ).encode("utf-8")


def json_loads(obj: Union[bytes, str]) -> TJSON:
    """Emulate orjson."""
    if isinstance(obj, bytes):
        obj = obj.decode("utf-8")
    return loads(obj)


with suppress(ImportError):
    from ujson import dumps as udumps
    from ujson import loads as json_loads  # type: ignore[assignment]

    def json_dumps(content) -> bytes:
        """Emulate orjson."""
        return udumps(content, ensure_ascii=False).encode("utf-8")


with suppress(ImportError):
    from orjson import dumps as json_dumps  # type: ignore[assignment,no-redef]
    from orjson import loads as json_loads  # type: ignore[assignment,no-redef]



"""Setup the library."""
from __future__ import annotations

import os
import sys

from setuptools import setup

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None


NO_EXTENSIONS = (
    sys.implementation.name != "cpython"
    or bool(os.environ.get("ASGI_TOOLS_NO_EXTENSIONS"))
    or cythonize is None
)

print("*********************")
print("* Pure Python build *" if NO_EXTENSIONS else "* Accelerated build *")
print("*********************")

setup(
    setup_requires=["wheel"],
    ext_modules=(
        []
        if NO_EXTENSIONS or cythonize is None
        else cythonize("asgi_tools/*.pyx", language_level=3)
    ),
)

# ruff: noqa: T201



from __future__ import annotations

from typing import ClassVar

from http_router import Router as HTTPRouter

from .errors import ASGIError, ASGIInvalidMethodError, ASGINotFoundError


class Router(HTTPRouter):
    """Rebind router errors."""

    NotFoundError: ClassVar[type[Exception]] = ASGINotFoundError
    RouterError: ClassVar[type[Exception]] = ASGIError
    InvalidMethodError: ClassVar[type[Exception]] = ASGIInvalidMethodError



from __future__ import annotations


async def test_aio_sleep():
    from asgi_tools._compat import aio_sleep

    await aio_sleep(1e-2)


async def test_aio_spawn():
    from asgi_tools._compat import aio_sleep, aio_spawn

    side_effects = {}

    async def task(name):
        side_effects[name] = True

    async with aio_spawn(task, "tests"):
        await aio_sleep(1e-2)

    assert side_effects["tests"] is True


async def test_aio_wait():
    from asgi_tools._compat import FIRST_COMPLETED, aio_sleep, aio_wait

    async def task(name, time):
        await aio_sleep(time)
        return name

    assert await aio_wait(task("t1", 2e-2), task("t2", 1e-2), strategy=FIRST_COMPLETED) == "t2"



"""ASGI responses."""

from __future__ import annotations

from email.utils import formatdate
from enum import Enum
from functools import partial
from hashlib import md5
from http import HTTPStatus
from http.cookies import SimpleCookie
from mimetypes import guess_type
from pathlib import Path
from stat import S_ISDIR
from typing import TYPE_CHECKING, Any, AsyncGenerator, Callable, Mapping, TypeVar
from urllib.parse import quote, quote_plus

from multidict import MultiDict

from ._compat import FIRST_COMPLETED, aio_stream_file, aio_wait, json_dumps
from .constants import BASE_ENCODING, DEFAULT_CHARSET
from .errors import ASGIConnectionClosedError, ASGIError
from .request import Request

if TYPE_CHECKING:
    from .types import TASGIMessage, TASGIReceive, TASGIScope, TASGISend

T = TypeVar("T")
TContent = TypeVar("TContent", str, bytes, dict, list, None)


class Response:
    """Base class for creating HTTP responses.

    :param content: A response's body
    :type content: str | bytes | dict | list | None
    :param status_code: An HTTP status code
    :type status_code: int
    :param headers: A dictionary of HTTP headers
    :type headers: dict[str, str]
    :param content_type: A string with the content-type
    :type content_type: str
    :param cookies: An initial dictionary of cookies
    :type cookies: dict[str, str]
    """

    headers: MultiDict  #: Multidict of response's headers
    cookies: SimpleCookie
    """ Set/Update cookies

    * `response.cookies[name] = value` ``str`` -- set a cookie's value
    * `response.cookies[name]['path'] = value` ``str`` -- set a cookie's path
    * `response.cookies[name]['expires'] = value` ``int`` -- set a cookie's expire
    * `response.cookies[name]['domain'] = value` ``str`` -- set a cookie's domain
    * `response.cookies[name]['max-age'] = value` ``int`` -- set a cookie's max-age
    * `response.cookies[name]['secure'] = value` ``bool``-- is the cookie
      should only be sent if request is SSL
    * `response.cookies[name]['httponly'] = value` ``bool`` -- is the cookie
      should be available through HTTP request only (not from JS)
    * `response.cookies[name]['samesite'] = value` ``str`` -- set a cookie's
      strategy ('lax'|'strict'|'none')

    """
    content_type: str | None = None
    status_code: int = HTTPStatus.OK.value

    def __init__(
        self,
        content: TContent,
        *,
        status_code: int | None = None,
        content_type: str | None = None,
        headers: dict[str, str] | None = None,
        cookies: dict[str, str] | None = None,
    ):
        """Setup the response."""
        self.content = self.process_content(content)
        self.headers: MultiDict = MultiDict(headers or {})
        self.cookies: SimpleCookie = SimpleCookie(cookies)
        if status_code is not None:
            self.status_code = status_code

        content_type = content_type or self.content_type
        if content_type:
            self.headers.setdefault(
                "content-type",
                (content_type.startswith("text/") and f"{content_type}; charset={DEFAULT_CHARSET}")
                or content_type,
            )

    def __str__(self) -> str:
        """Stringify the response."""
        return f"{self.status_code}"

    def __repr__(self) -> str:
        """Stringify the response."""
        return f"<{ self.__class__.__name__ } '{ self }'>"

    async def __call__(self, _, __, send: TASGISend):
        """Behave as an ASGI application."""
        self.headers.setdefault("content-length", str(len(self.content)))

        await send(self.msg_start())
        await send({"type": "http.response.body", "body": self.content})

    @staticmethod
    def process_content(content: TContent) -> bytes:
        """Process content into bytes."""
        if not isinstance(content, bytes):
            return str(content).encode(DEFAULT_CHARSET)
        return content

    def msg_start(self) -> TASGIMessage:
        """Get ASGI response start message."""
        headers = [
            (key.encode(BASE_ENCODING), str(val).encode(BASE_ENCODING))
            for key, val in self.headers.items()
        ]

        for cookie in self.cookies.values():
            headers = [
                *headers,
                (b"set-cookie", cookie.output(header="").strip().encode(BASE_ENCODING)),
            ]

        return {
            "type": "http.response.start",
            "status": self.status_code,
            "headers": headers,
        }


class ResponseText(Response):
    """Returns plain text responses (text/plain)."""

    content_type = "text/plain"


class ResponseHTML(Response):
    """Returns HTML responses (text/html)."""

    content_type = "text/html"


class ResponseJSON(Response):
    """Returns JSON responses (application/json).

    The class optionally supports `ujson <https://pypi.org/project/ujson/>`_ and `orjson
    <https://pypi.org/project/orjson/>`_ JSON libraries. Install one of them to use instead
    the standard library.

    """

    content_type = "application/json"

    @staticmethod
    def process_content(content) -> bytes:
        """Dumps the given content."""
        return json_dumps(content)


class ResponseStream(Response):
    """Streams response body as chunks.

    :param content: An async generator to stream the response's body
    :type content: AsyncGenerator
    """

    def __init__(self, stream: AsyncGenerator[Any, None], **kwargs):
        super().__init__(b"", **kwargs)
        self.stream = stream

    async def listen_for_disconnect(self, receive: TASGIReceive):
        """Listen for the client has been disconnected."""
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                break

    async def stream_response(self, send: TASGISend):
        """Stream response content."""
        await send(self.msg_start())
        async for chunk in self.stream:
            await send(
                {
                    "type": "http.response.body",
                    "body": self.process_content(chunk),
                    "more_body": True,
                },
            )

        await send({"type": "http.response.body", "body": b""})

    async def __call__(self, _, receive, send: TASGISend) -> None:
        """Behave as an ASGI application."""
        await aio_wait(
            self.listen_for_disconnect(receive),
            self.stream_response(send),
            strategy=FIRST_COMPLETED,
        )


class ResponseSSE(ResponseStream):
    """Streams Server-Sent Events (SSE).

    :param content: An async generator to stream the events
    :type content: AsyncGenerator
    """

    content_type = "text/event-stream"

    def msg_start(self) -> TASGIMessage:
        """Set cache-control header."""
        self.headers.setdefault("Cache-Control", "no-cache")
        return super().msg_start()

    @staticmethod
    def process_content(content) -> bytes:
        """Prepare a chunk from stream generator to send."""
        if isinstance(content, dict):
            content = "\n".join(f"{k}: {v}" for k, v in content.items())

        if not isinstance(content, bytes):
            content = content.encode(DEFAULT_CHARSET)

        return content + b"\n\n"


class ResponseFile(ResponseStream):
    """Serves files as HTTP responses.

    :param filepath: The filepath to the file
    :type filepath: str | Path
    :param chunk_size: Default chunk size (32768)
    :type chunk_size: int
    :param filename: If set, `Content-Disposition` header will be generated
    :type filename: str
    :param headers_only: Return only file headers
    :type headers_only: bool

    """

    def __init__(
        self,
        filepath: str | Path,
        *,
        chunk_size: int = 64 * 1024,
        filename: str | None = None,
        headers_only: bool = False,
        **kwargs,
    ) -> None:
        """Store filepath to self."""
        try:
            stat = Path(filepath).stat()
        except FileNotFoundError as exc:
            raise ASGIError(*exc.args) from exc

        if S_ISDIR(stat.st_mode):
            raise ASGIError(f"It's a directory: {filepath}")

        super().__init__(
            empty() if headers_only else aio_stream_file(filepath, chunk_size),
            **kwargs,
        )

        headers = self.headers
        if filename and "content-disposition" not in headers:
            headers["content-disposition"] = f"attachment; filename*=UTF-8''{quote(filename)}"

        if "content-type" not in headers:
            headers["content-type"] = guess_type(filename or str(filepath))[0] or "text/plain"

        headers.setdefault("content-length", str(stat.st_size))
        headers.setdefault("last-modified", formatdate(stat.st_mtime, usegmt=True))
        etag = str(stat.st_mtime) + "-" + str(stat.st_size)
        headers.setdefault("etag", md5(etag.encode()).hexdigest())  # noqa: S324


class ResponseWebSocket(Response):
    """Provides a WebSocket handler interface."""

    class STATES(Enum):
        """Represent websocket states."""

        CONNECTING = 0
        CONNECTED = 1
        DISCONNECTED = 2

    def __init__(
        self,
        scope: TASGIScope,
        receive: TASGIReceive | None = None,
        send: TASGISend | None = None,
    ) -> None:
        """Initialize the websocket response."""
        if isinstance(scope, Request):
            receive, send = scope.receive, scope.send

        if not receive or not send:
            raise ASGIError("Invalid initialization")

        super().__init__(b"")
        self._receive: TASGIReceive = receive
        self._send: TASGISend = send
        self.state = self.STATES.CONNECTING
        self.partner_state = self.STATES.CONNECTING

    async def __call__(self, _, __, send: TASGISend):
        """Close websocket if the response has been returned."""
        await send({"type": "websocket.close"})

    async def __aenter__(self):
        """Use it as async context manager."""
        await self.accept()
        return self

    async def __aexit__(self, *_):
        """Exit async context."""
        await self.close()

    @property
    def connected(self) -> bool:
        """Check that is the websocket connected."""
        return self.state == self.partner_state == self.STATES.CONNECTED

    async def _connect(self) -> bool:
        """Wait for connect message."""
        if self.partner_state == self.STATES.CONNECTING:
            msg = await self._receive()
            assert msg.get("type") == "websocket.connect"
            self.partner_state = self.STATES.CONNECTED

        return self.partner_state == self.STATES.CONNECTED

    async def accept(self, **params) -> None:
        """Accept a websocket connection."""
        if self.partner_state == self.STATES.CONNECTING:
            await self._connect()

        await self.send({"type": "websocket.accept", **params})
        self.state = self.STATES.CONNECTED

    async def close(self, code: int = 1000) -> None:
        """Sent by the application to tell the server to close the connection."""
        if self.connected:
            await self.send({"type": "websocket.close", "code": code})
        self.state = self.STATES.DISCONNECTED

    async def send(self, msg: dict | str | bytes, msg_type="websocket.send") -> None:
        """Send the given message to a client."""
        if self.state == self.STATES.DISCONNECTED:
            raise ASGIConnectionClosedError

        if not isinstance(msg, dict):
            msg = {"type": msg_type, ((isinstance(msg, str) and "text") or "bytes"): msg}

        return await self._send(msg)

    async def send_json(self, data) -> None:
        """Serialize the given data to JSON and send to a client."""
        return await self._send({"type": "websocket.send", "bytes": json_dumps(data)})

    async def receive(self, *, raw: bool = False) -> TASGIMessage | str:
        """Receive messages from a client.

        :param raw: Receive messages as is.
        """
        if self.partner_state == self.STATES.DISCONNECTED:
            raise ASGIConnectionClosedError

        if self.partner_state == self.STATES.CONNECTING:
            await self._connect()
            return await self.receive(raw=raw)

        msg = await self._receive()
        if msg["type"] == "websocket.disconnect":
            self.partner_state = self.STATES.DISCONNECTED

        return msg if raw else parse_websocket_msg(msg, charset=DEFAULT_CHARSET)


class ResponseRedirect(Response, BaseException):
    """Creates HTTP redirects. Uses a 307 status code by default.

    :param url: A string with the new location
    :type url: str
    """

    status_code: int = HTTPStatus.TEMPORARY_REDIRECT.value

    def __init__(self, url: str, status_code: int | None = None, **kwargs) -> None:
        """Set status code and prepare location."""
        super().__init__(b"", status_code=status_code, **kwargs)
        assert (
            300 <= self.status_code < 400
        ), f"Invalid status code for redirection: {self.status_code}"
        self.headers["location"] = quote_plus(url, safe=":/%#?&=@[]!$&'()*+,;")


class ResponseErrorMeta(type):
    """Generate Response Errors by HTTP names."""

    # TODO: From python 3.9 -> partial['ResponseError]
    def __getattr__(cls, name: str) -> Callable[..., ResponseError]:
        """Generate Response Errors by HTTP names."""
        status = HTTPStatus[name]
        return partial(
            lambda *args, **kwargs: cls(*args, **kwargs),
            status_code=status.value,
        )


class ResponseError(Response, BaseException, metaclass=ResponseErrorMeta):
    """Helper for returning HTTP errors. Uses a 500 status code by default.

    :param message: A string with the error's message (HTTPStatus messages will be used by default)
    :type message: str

    You able to use :py:class:`http.HTTPStatus` properties with the `ResponseError` class

    .. code-block:: python

        response = ResponseError.BAD_REQUEST('invalid data')
        response = ResponseError.NOT_FOUND()
        response = ResponseError.BAD_GATEWAY()
        # and etc

    """

    status_code: int = HTTPStatus.INTERNAL_SERVER_ERROR.value

    # Typing annotations
    if TYPE_CHECKING:
        BAD_REQUEST: Callable[..., ResponseError]  # 400
        UNAUTHORIZED: Callable[..., ResponseError]  # 401
        PAYMENT_REQUIRED: Callable[..., ResponseError]  # 402
        FORBIDDEN: Callable[..., ResponseError]  # 403
        NOT_FOUND: Callable[..., ResponseError]  # 404
        METHOD_NOT_ALLOWED: Callable[..., ResponseError]  # 405
        NOT_ACCEPTABLE: Callable[..., ResponseError]  # 406
        PROXY_AUTHENTICATION_REQUIRED: Callable[..., ResponseError]  # 407
        REQUEST_TIMEOUT: Callable[..., ResponseError]  # 408
        CONFLICT: Callable[..., ResponseError]  # 409
        GONE: Callable[..., ResponseError]  # 410
        LENGTH_REQUIRED: Callable[..., ResponseError]  # 411
        PRECONDITION_FAILED: Callable[..., ResponseError]  # 412
        REQUEST_ENTITY_TOO_LARGE: Callable[..., ResponseError]  # 413
        REQUEST_URI_TOO_LONG: Callable[..., ResponseError]  # 414
        UNSUPPORTED_MEDIA_TYPE: Callable[..., ResponseError]  # 415
        REQUESTED_RANGE_NOT_SATISFIABLE: Callable[..., ResponseError]  # 416
        EXPECTATION_FAILED: Callable[..., ResponseError]  # 417
        # TODO: From python 3.9
        # IM_A_TEAPOT: Callable[..., ResponseError]                       # 418
        # MISDIRECTED_REQUEST: Callable[..., ResponseError]               # 421
        UNPROCESSABLE_ENTITY: Callable[..., ResponseError]  # 422
        LOCKED: Callable[..., ResponseError]  # 423
        FAILED_DEPENDENCY: Callable[..., ResponseError]  # 424
        TOO_EARLY: Callable[..., ResponseError]  # 425
        UPGRADE_REQUIRED: Callable[..., ResponseError]  # 426
        PRECONDITION_REQUIRED: Callable[..., ResponseError]  # 428
        TOO_MANY_REQUESTS: Callable[..., ResponseError]  # 429
        REQUEST_HEADER_FIELDS_TOO_LARGE: Callable[..., ResponseError]  # 431
        # TODO: From python 3.9
        # UNAVAILABLE_FOR_LEGAL_REASONS: Callable[..., ResponseError]     # 451

        INTERNAL_SERVER_ERROR: Callable[..., ResponseError]  # 500
        NOT_IMPLEMENTED: Callable[..., ResponseError]  # 501
        BAD_GATEWAY: Callable[..., ResponseError]  # 502
        SERVICE_UNAVAILABLE: Callable[..., ResponseError]  # 503
        GATEWAY_TIMEOUT: Callable[..., ResponseError]  # 504
        HTTP_VERSION_NOT_SUPPORTED: Callable[..., ResponseError]  # 505
        VARIANT_ALSO_NEGOTIATES: Callable[..., ResponseError]  # 506
        INSUFFICIENT_STORAGE: Callable[..., ResponseError]  # 507
        LOOP_DETECTED: Callable[..., ResponseError]  # 508
        NOT_EXTENDED: Callable[..., ResponseError]  # 510
        NETWORK_AUTHENTICATION_REQUIRED: Callable[..., ResponseError]  # 511

    def __init__(self, message=None, status_code: int | None = None, **kwargs):
        """Check error status."""
        content = message or HTTPStatus(status_code or self.status_code).description
        super().__init__(content=content, status_code=status_code, **kwargs)
        assert self.status_code >= 400, f"Invalid status code for an error: {self.status_code}"


CAST_RESPONSE: Mapping[type, type[Response]] = {
    bool: ResponseJSON,
    bytes: ResponseHTML,
    dict: ResponseJSON,
    int: ResponseJSON,
    list: ResponseJSON,
    str: ResponseHTML,
    type(None): ResponseJSON,
}


def parse_response(response, headers: dict | None = None) -> Response:
    """Parse the given object and convert it into a asgi_tools.Response."""
    if isinstance(response, Response):
        return response

    rtype = type(response)
    response_type = CAST_RESPONSE.get(rtype)
    if response_type:
        return response_type(response, headers=headers)

    if rtype is tuple:
        status, *contents = response
        assert isinstance(status, int), "Invalid Response Status"
        if len(contents) > 1:
            headers, *contents = contents
        response = parse_response(
            contents[0] or "" if contents else "",
            headers=headers,
        )
        response.status_code = status
        return response

    return ResponseText(str(response), headers=headers)


def parse_websocket_msg(msg: TASGIMessage, charset: str | None = None) -> TASGIMessage | str:
    """Prepare websocket message."""
    data = msg.get("text")
    if data:
        return data

    data = msg.get("bytes")
    if data:
        return data.decode(charset)

    return msg


async def empty():
    yield b""


# ruff: noqa: ERA001



from __future__ import annotations

import asyncio
from multiprocessing import Value

import pytest

from asgi_tools._compat import FIRST_COMPLETED, aio_sleep, aio_timeout, aio_wait


async def test_aio_sleep():
    await aio_sleep(1e-2)


async def test_aio_timeout_disabled():
    async with aio_timeout(0):
        await aio_sleep(1e-2)


async def test_aio_timeout():
    with pytest.raises((TimeoutError, asyncio.TimeoutError)):  # python 310
        async with aio_timeout(1e-2):
            await aio_sleep(1)


async def coro(num, exc=None):
    await aio_sleep(1e-2 * num)
    if exc:
        raise exc
    return num


async def test_aio_wait():
    results = await aio_wait(coro(3), coro(2), coro(1))
    assert sorted(results) == [1, 2, 3]


async def test_aio_wait_first_completed():
    result = await aio_wait(coro(3), coro(2), coro(1), strategy=FIRST_COMPLETED)
    assert result == 1


def test_compat_json():
    from asgi_tools._compat import json_dumps, json_loads

    data = json_dumps({"test": 42})
    assert data
    assert isinstance(data, bytes)

    data = json_loads(data)
    assert data
    assert data == {"test": 42}
